# mypy: ignore-errors
# TODO: type this file
import math
from typing import Optional
from typing import Tuple

import gym
import torch
from einops import rearrange
from loguru import logger
from torch import Tensor
from torch import nn
from torch.distributions import Bernoulli
from torch.distributions import Distribution
from torch.distributions import Independent
from torch.distributions import Normal
from torch.nn import functional as F
from tree import map_structure

from avalon.agent.common.action_model import DictActionHead
from avalon.agent.common.models import ACTIVATION_MODULE_LOOKUP
from avalon.agent.common.models import MLP
from avalon.agent.common.models import ActivationFunction
from avalon.agent.common.types import ActionBatch
from avalon.agent.common.types import LatentBatch
from avalon.agent.common.types import ObservationBatch
from avalon.agent.dreamer.tools import static_scan


def init_weights(m: torch.nn.Module, gain: float = 1.0) -> None:
    # It appears that tf uses glorot init with gain 1 by default for all weights, and zero for all biases
    # at least for these layers
    if isinstance(m, nn.Linear):
        # This is the default used in a tensorflow keras Dense layer, with gain=1
        torch.nn.init.xavier_uniform_(m.weight, gain=gain)
        m.bias.data.fill_(0.0)
    if isinstance(m, nn.Conv2d):
        torch.nn.init.xavier_uniform_(m.weight, gain=gain)
        m.bias.data.fill_(0.0)
    if isinstance(m, nn.ConvTranspose2d):
        torch.nn.init.xavier_uniform_(m.weight, gain=gain)
        m.bias.data.fill_(0.0)


class GRUCell(nn.Module):
    # This is the custom cell Danijar wrote in his dreamerv2 repo.
    # Default initializer values are the values used for all configs.
    def __init__(
        self, input_dim: int, state_size: int, norm: bool = True, act=torch.tanh, update_bias: float = -1
    ) -> None:
        """
        - input_dim: the size of the input vector
        - state_size: the size of the reset, cand, update, state, and output vectors
        """
        super().__init__()
        self._state_size = state_size
        self._act = act
        self._update_bias = update_bias
        self._layer = torch.nn.Linear(
            in_features=input_dim + state_size, out_features=3 * state_size, bias=norm is not None
        )
        if norm:
            self._norm = torch.nn.LayerNorm(normalized_shape=3 * state_size)

        self.apply(init_weights)

    def forward(self, inputs, state):
        # inputs should be hidden_size, state should be deter_size
        parts = self._layer(torch.cat([inputs, state], -1))
        if self._norm:
            parts = self._norm(parts)
        reset, cand, update = torch.chunk(parts, 3, -1)
        reset = torch.sigmoid(reset)
        cand = self._act(reset * cand)
        update = torch.sigmoid(update + self._update_bias)
        output = update * cand + (1 - update) * state
        return output


class RSSM(nn.Module):
    def __init__(self, actdim, embed_size=1024, stoch=30, deter=200, hidden=200, act=F.elu) -> None:
        super().__init__()
        self._activation = act
        self._stoch_size = stoch
        self._deter_size = deter
        self._hidden_size = hidden
        self.min_std = 0.1  # this is the value used in the dreamerv2 configs

        # These two layers are common/shared for the prior and posterior.
        self.img1 = nn.Linear(actdim + stoch, self._hidden_size)
        self._cell = GRUCell(input_dim=self._hidden_size, state_size=self._deter_size, norm=True)

        # Prior model
        self.img2 = nn.Linear(self._deter_size, self._hidden_size)
        self.img3 = nn.Linear(self._hidden_size, 2 * self._stoch_size)

        # Posterior model
        self._embed_size = embed_size
        self.obs1 = nn.Linear(self._deter_size + self._embed_size, self._hidden_size)
        self.obs2 = nn.Linear(self._hidden_size, 2 * self._stoch_size)

        self.apply(init_weights)

    def initial(self, batch_size: int, device) -> LatentBatch:
        """This is the initial latent state."""
        return dict(
            mean=torch.zeros([batch_size, self._stoch_size], device=device),
            std=torch.zeros([batch_size, self._stoch_size], device=device),
            stoch=torch.zeros([batch_size, self._stoch_size], device=device),
            deter=torch.zeros([batch_size, self._deter_size], device=device),
        )

    def observe(self, embed: Tensor, action: ActionBatch, state=None) -> Tuple[LatentBatch, LatentBatch]:
        """Generates state estimations given a sequence of observations and actions.
        Only used in training?"""

        if state is None:
            state = self.initial(batch_size=embed.shape[0], device=embed.device)
        # these are moving the time axis to the front, for the static_scan
        embed = rearrange(embed, "b t ... -> t b ...")
        action = {k: rearrange(v, "b t ... -> t b ...") for k, v in action.items()}
        # here the nest structure is a tuple of tensors, and a tuple of dicts of tensors
        # (state, state) is just to give the proper structure for the output
        post, prior = static_scan(
            lambda prev, inputs: self.obs_step(prev[0], *inputs), (action, embed), (state, state)
        )
        # Moving the batch axis back in front
        post = {k: v.permute(1, 0, 2) for k, v in post.items()}
        prior = {k: v.permute(1, 0, 2) for k, v in prior.items()}
        return post, prior

    def imagine(self, action: ActionBatch, state: LatentBatch) -> LatentBatch:
        # TODO: only used for the imagination viz. figure out how to share with train() imagine.
        # or probably just move this into the viz logic
        # Takes a list of actions and an initial state and rolls it out in imagination.
        assert isinstance(state, dict), state
        prior = static_scan(self.img_step, action, state)
        prior = {k: v.permute(1, 0, 2) for k, v in prior.items()}
        return prior

    def get_feat(self, state: LatentBatch) -> Tensor:
        """Combine stoch and deter state into a single vector."""
        feat = torch.cat([state["stoch"], state["deter"]], -1)
        return feat

    def get_dist(self, state: LatentBatch) -> torch.distributions.Distribution:
        dist = Independent(Normal(state["mean"], state["std"]), 1)
        return dist

    def obs_step(
        self, prev_state: LatentBatch, prev_action: ActionBatch, embed: Tensor
    ) -> Tuple[LatentBatch, LatentBatch]:
        """Observation is the process of taking s_{t-1}, a_{t-1}, and o_{t} and generating s_{t}.
        In other words, estimating the current (unobserved) state given the current observation.

        This is what the posterior model does."""
        prior = self.img_step(prev_state, prev_action)

        # Compute the posterior
        # embed is the observation embedding
        x = torch.cat([prior["deter"], embed], -1)
        x = self._activation(self.obs1(x))
        x = self.obs2(x)
        mean, std = torch.chunk(x, 2, -1)

        std = 2 * torch.sigmoid(std / 2) + self.min_std
        stoch_dist = self.get_dist({"mean": mean, "std": std})

        # TODO: this should be the mode if we're doing eval. Does eval ever use this codepath?
        stoch = stoch_dist.rsample()
        post = {"mean": mean, "std": std, "stoch": stoch, "deter": prior["deter"]}
        return post, prior

    def img_step(self, prev_state: LatentBatch, prev_action: ActionBatch) -> LatentBatch:
        """Imagination is the process of taking s_{t-1} and a_{t-1} and generating s_{t}.
        In other words, imagining what the next state will be given the previous state and action.

        This is what the prior model does."""
        # Compute the RNN (this is the deterministic part)
        # Flatten actions out into a vector per batch element. Batch must always only have 1 axis here.
        prev_action = torch.cat([x.reshape([x.shape[0], -1]) for x in prev_action.values()], dim=-1)
        x = torch.cat([prev_state["stoch"], prev_action], -1)
        x = self._activation(self.img1(x))
        deter = self._cell(x, prev_state["deter"])

        # Compute the prior
        y = self._activation(self.img2(deter))
        y = self.img3(y)
        mean, std = torch.chunk(y, 2, -1)
        std = 2 * torch.sigmoid(std / 2) + self.min_std
        stoch = self.get_dist({"mean": mean, "std": std}).rsample()
        prior = {"mean": mean, "std": std, "stoch": stoch, "deter": deter}
        return prior

    def kl_loss(
        self, post: LatentBatch, prior: LatentBatch, balance: float = 0.8, free: float = 0.0
    ) -> Tuple[Tensor, Tensor]:
        kld = torch.distributions.kl_divergence
        sg = lambda x: map_structure(torch.detach, x)
        if math.isclose(balance, 0.5):
            value = kld(self.get_dist(post), self.get_dist(prior))
            # Shape of value is (batch, timesteps)
            loss = torch.clamp(value, min=free).mean()
        else:
            value_lhs = value = kld(self.get_dist(post), self.get_dist(sg(prior)))
            value_rhs = kld(self.get_dist(sg(post)), self.get_dist(prior))
            loss_lhs = torch.clamp(value_lhs.mean(), min=free)
            loss_rhs = torch.clamp(value_rhs.mean(), min=free)
            loss = (1 - balance) * loss_lhs + balance * loss_rhs
        return loss, value


def is_image_space(x: gym.spaces.Space) -> bool:
    return isinstance(x, gym.spaces.Box) and len(x.shape) == 3


def is_vector_space(x: gym.spaces.Space) -> bool:
    return isinstance(x, gym.spaces.Box) and len(x.shape) == 1


class HybridEncoder(nn.Module):
    """Takes a dict obs space composed of image and vector Box spaces, and encodes it all to a single latent vector."""

    def __init__(
        self,
        obs_space: gym.spaces.Dict,
        mlp_hidden_dim=400,
        activation_function: ActivationFunction = ActivationFunction.ACTIVATION_ELU,
    ) -> None:
        super().__init__()
        assert isinstance(obs_space, gym.spaces.Dict)
        self.obs_space = obs_space
        # Out dim is dynamic - just the native output size of the MLP + conv, concatted together.
        self.out_dim = 0
        activation_module = ACTIVATION_MODULE_LOOKUP[activation_function.value]

        self.image_keys = [k for k, v in obs_space.spaces.items() if is_image_space(v)]
        self.vector_keys = [k for k, v in obs_space.spaces.items() if is_vector_space(v)]
        # TODO: handle scalar spaces! Either here or coerce them to vectors in the wrapper.
        scalar_spaces = [k for k, v in obs_space.spaces.items() if v.shape == ()]
        if len(scalar_spaces) > 0:
            logger.warning("reminder: ignoring scalar spaces!")
        assert len(obs_space) == len(self.image_keys) + len(self.vector_keys) + len(scalar_spaces)

        if self.image_keys:
            # We expect images to be (c, h, w)
            img_size = obs_space.spaces[self.image_keys[0]].shape[1]
            self.img_channels = 0
            for key in self.image_keys:
                space = obs_space.spaces[key]
                assert space.shape[1] == img_size and space.shape[2] == img_size
                self.img_channels += space.shape[0]

            self.image_encoder = ConvEncoder(self.img_channels, img_size)
            self.cnn_dim = self.image_encoder.output_dim
            self.out_dim += self.cnn_dim
            assert 100 < self.cnn_dim < 2000, "just a good heuristic"

        if self.vector_keys:
            self.vector_dim = 0
            for key in self.vector_keys:
                space = obs_space.spaces[key]
                self.vector_dim += space.shape[0]

            self.vector_encoder_fc = MLP(
                input_dim=self.vector_dim,
                hidden_dim=mlp_hidden_dim,
                output_dim=mlp_hidden_dim,
                num_layers=4,
                activation_function=activation_function,
            )
            self.vector_activation = activation_module()
            self.out_dim += mlp_hidden_dim

        self.apply(init_weights)

    def forward(self, obs: ObservationBatch) -> Tensor:
        encodings = []
        if self.image_keys:
            image_parts = [obs[key] for key in self.image_keys]
            # Ensure that we applied the post-transform properly everywhere
            assert all([x.dtype == torch.float32 for x in image_parts])
            image_obs = torch.cat(image_parts, dim=-3)
            img_encoding = self.image_encoder(image_obs)
            assert img_encoding.shape[-1] == self.cnn_dim
            encodings.append(img_encoding)

        if self.vector_keys:
            vector_parts = [obs[key] for key in self.vector_keys]
            vector_obs = torch.cat(vector_parts, dim=-1)
            assert vector_obs.shape[-1] == self.vector_dim
            x = self.vector_encoder_fc(vector_obs)
            x = self.vector_activation(x)
            encodings.append(x)

        encoding = torch.cat(encodings, dim=-1)
        assert encoding.shape[-1] == self.out_dim
        return encoding


class HybridDecoder(nn.Module):
    """The inverse of the HybridEncoder"""

    def __init__(
        self,
        obs_space: gym.spaces.Dict,
        latent_dim: int,
        skip_keys: Tuple[str] = (),
        mlp_hidden_dim=400,
        activation_function: ActivationFunction = ActivationFunction.ACTIVATION_ELU,
    ) -> None:
        super().__init__()
        assert isinstance(obs_space, gym.spaces.Dict)
        self.obs_space = obs_space

        self.image_keys = [k for k, v in obs_space.spaces.items() if is_image_space(v) if k not in skip_keys]
        self.vector_keys = [k for k, v in obs_space.spaces.items() if is_vector_space(v) if k not in skip_keys]
        # TODO: handle scalar spaces! Either here or coerce them to vectors in the wrapper.
        scalar_spaces = [k for k, v in obs_space.spaces.items() if v.shape == ()]
        skipped_keys = [k for k in obs_space.spaces.keys() if k in skip_keys]
        assert len(obs_space) == len(self.image_keys) + len(self.vector_keys) + len(scalar_spaces) + len(skipped_keys)

        if self.image_keys:
            # We expect images to be (c, h, w)
            img_size = obs_space.spaces[self.image_keys[0]].shape[1]
            self.img_channels = 0
            for key in self.image_keys:
                space = obs_space.spaces[key]
                assert space.shape[1] == img_size and space.shape[2] == img_size
                self.img_channels += space.shape[0]
            # The decoder already has a linear to adapt the input size
            self.image_decoder = ConvDecoder(input_dim=latent_dim, out_channels=self.img_channels, res=img_size)

        if self.vector_keys:
            self.vector_dim = 0
            for key in self.vector_keys:
                space = obs_space.spaces[key]
                self.vector_dim += space.shape[0]
            self.vector_decoder_fc = MLP(
                input_dim=latent_dim,
                hidden_dim=mlp_hidden_dim,
                output_dim=self.vector_dim,
                num_layers=5,  # danijar has 4 hidden and 1 output layer (per space, which we have merged)
                activation_function=activation_function,
            )

        self.apply(init_weights)

    def forward(self, latent: Tensor) -> ObservationBatch:
        # We're only going to handle latents of shape (b, t, latent_dim) here
        assert len(latent.shape) == 3
        batch_size, timesteps = latent.shape[:2]
        out = {}
        if self.image_keys:
            decoded_img = self.image_decoder(latent)
            # assert decoded_img.shape == (batch_size, timesteps, self.img_channels, 64, 64)
            start_channel = 0
            for key in self.image_keys:
                channels = self.obs_space.spaces[key].shape[0]
                out[key] = decoded_img[:, :, start_channel : start_channel + channels]
                start_channel += channels

        if self.vector_keys:
            decoded_vec = self.vector_decoder_fc(latent)
            assert decoded_vec.shape == (batch_size, timesteps, self.vector_dim)
            start_dim = 0
            for key in self.vector_keys:
                dims = self.obs_space.spaces[key].shape[0]
                out[key] = decoded_vec[:, :, start_dim : start_dim + dims]
                start_dim += dims

        return out


def compute_conv_output_res(
    input_res: int, kernel_size: int, stride: int = 1, padding: int = 0, dilation: int = 1
) -> int:
    return math.floor(((input_res + 2 * padding - dilation * (kernel_size - 1) - 1) / stride) + 1)


def compute_conv_transpose_output_res(
    input_res: int, kernel_size: int, stride: int = 1, padding: int = 0, output_padding: int = 0, dilation: int = 1
) -> int:
    return (input_res - 1) * stride - 2 * padding + dilation * (kernel_size - 1) + output_padding + 1


class ConvEncoder(nn.Module):
    def __init__(self, input_channels: int = 3, input_res: int = 96) -> None:
        super().__init__()
        self.input_channels = input_channels
        self.input_res = input_res

        # for 64x64, we will output a shape of 2x2x384 (or whatever ordering we use here..)
        # for 96x96, we will output the same, requiring 1 extra layer. 2x2x384
        kernels = {
            64: [4, 4, 4, 4],
            96: [4, 4, 4, 4, 2],
        }[input_res]
        depth = {
            64: 48,
            96: 24,
        }[input_res]
        layers = []
        in_channels = input_channels
        current_res = input_res
        for i, kernel in enumerate(kernels):
            out_channels = 2**i * depth
            conv = nn.Conv2d(
                in_channels=in_channels, out_channels=out_channels, kernel_size=kernel, stride=2, padding=0
            )
            in_channels = out_channels
            current_res = compute_conv_output_res(current_res, kernel, stride=2, padding=0)
            layers.append(conv)
            layers.append(nn.ELU())

        self.conv = nn.Sequential(*layers)

        self.output_dim = in_channels * (current_res * current_res)
        self.apply(init_weights)

    def forward(self, obs: Tensor) -> Tensor:
        # This can get called with shape (b, c, h, w) or (b, t, c, h, w)
        x = obs
        # Convert (b, t, c, h, w) to (b, c, h, w) if starting with the former
        x = torch.reshape(x, (-1,) + x.shape[-3:])
        assert x.shape[1:] == (self.input_channels, self.input_res, self.input_res)
        x = self.conv(x)
        # Reshape back to the original b/t shape, flatten out the rest
        shape = obs.shape[:-3] + (self.output_dim,)
        x = torch.reshape(x, shape)
        return x


class ConvDecoder(nn.Module):
    def __init__(self, input_dim: int, out_channels: int, depth: int = 48, res: int = 96) -> None:
        super().__init__()
        self._depth = depth
        self.res = res
        self._shape = (out_channels, self.res, self.res)
        self.out_channels = out_channels

        self.embedding_size = 32 * self._depth
        self.fc1 = nn.Linear(input_dim, self.embedding_size)

        layers = []
        kernels = {
            64: [5, 5, 6, 6],
            96: [4, 4, 4, 5, 4],
        }[res]
        current_res = 1
        current_in_channels = self.embedding_size
        for i, kernel in enumerate(kernels):
            current_out_channels = 2 ** (len(kernels) - i - 2) * depth
            if i == len(kernels) - 1:
                conv = nn.ConvTranspose2d(current_in_channels, out_channels, kernel_size=kernel, stride=2)
                layers.append(conv)
            else:
                conv = nn.ConvTranspose2d(current_in_channels, current_out_channels, kernel_size=kernel, stride=2)
                layers.append(conv)
                layers.append(nn.ELU())
            current_in_channels = current_out_channels
            current_res = compute_conv_transpose_output_res(current_res, kernel, stride=2)

        assert current_res == res, (current_res, res)
        self.conv = nn.Sequential(*layers)

        self.apply(init_weights)

    def forward(self, features: Tensor) -> Tensor:
        # c is stoch + deter
        b, t, c = features.shape
        x = self.fc1(features)  # this is not followed by a nonlinearity
        x = rearrange(x, "b t c -> (b t) c 1 1", b=b, t=t, c=self.embedding_size)
        x = self.conv(x)
        dist_mean = rearrange(x, "(b t) c h w-> b t c h w", b=b, t=t, h=self.res, w=self.res, c=self.out_channels)
        # shape is [50, 50, self.res, self.res, 3]
        return dist_mean


class DenseDecoder(nn.Module):
    def __init__(
        self,
        output_shape: Tuple[int, ...],
        layers: int,
        in_dim: int,
        units: int,
        dist: str,
        action_head: Optional[DictActionHead] = None,
        activation_function: ActivationFunction = ActivationFunction.ACTIVATION_ELU,
    ) -> None:
        super().__init__()
        self.output_shape = output_shape
        self.dist = dist

        # Dreamerv2 has an additional Linear layer in the `DistLayer`.
        # So they actually have `layers` + 1 Linear layers.
        self.mlp = MLP(
            in_dim,
            hidden_dim=units,
            output_dim=math.prod(output_shape),
            num_layers=layers + 1,
            activation_function=activation_function,
        )
        if self.dist == "action_head":
            assert action_head
            self.action_head = action_head

        self.apply(init_weights)

    def forward(self, x: Tensor, raw_feats=False) -> Distribution:
        batch_shape = x.shape[:-1]
        x = self.mlp(x)
        x = x.view(batch_shape + self.output_shape)
        if raw_feats:
            return x
        if self.dist == "normal":
            # The std 1 here is kinda weird. Why not let it predict its own std?
            # Oh, this is essentially just a MSE loss with this configuration.
            return Independent(Normal(x, 1), len(self.output_shape))
        elif self.dist == "binary":
            # x is just the batch shape, so this will properly construct a dist with proper batch shape and scalar event shape
            return Bernoulli(logits=x)
        elif self.dist == "action_head":
            dist = self.action_head(x)
            return dist
        raise NotImplementedError(self.dist)



from typing import List

import attr
import numpy as np
from godot_parser import Node as GDNode

from avalon.datagen.world_creation.entities.entity import Entity
from avalon.datagen.world_creation.types import GodotScene
from avalon.datagen.world_creation.types import Point3DNP


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DoorLock(Entity):
    entity_id: int = -1

    def get_additional_door_body_nodes(
        self, scene: GodotScene, body_size: np.ndarray, body_centroid: Point3DNP = np.array([0, 0, 0])
    ) -> List[GDNode]:
        return []



from random import Random
from typing import Dict
from typing import Generic
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import attr

from avalon.contrib.serialization import Serializable
from avalon.datagen.data_config import AbstractDataConfig
from avalon.datagen.data_config import AbstractRange


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class IntRange(AbstractRange):
    min_ge: int
    max_lt: int

    @staticmethod
    def n(n: int) -> "IntRange":
        return IntRange(0, n)

    @property
    def size(self) -> int:
        return self.max_lt - self.min_ge

    @property
    def midpoint(self) -> float:
        return self.min_ge + self.size / 2.0

    def contains(self, value: float) -> bool:
        return self.min_ge <= value < self.max_lt

    def overlap(self, other: "IntRange") -> Optional["IntRange"]:
        min_ge = max(self.min_ge, other.min_ge)
        max_lt = min(self.max_lt, other.max_lt)
        if min_ge >= max_lt:
            return None
        return IntRange(min_ge, max_lt)


# TODO: this was missing, add it and verify nothing breaks
# @attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class DataConfigImplementation(AbstractDataConfig):
    video_min: int
    video_max: int
    frame_max: int
    dir_root: str
    random_key: str

    output_file_name_format: str
    is_using_shared_caches: bool
    is_generating_paired_videos: bool

    def get_is_using_shared_caches(self) -> bool:
        return self.is_using_shared_caches

    def set_is_using_shared_caches(
        self: "DataConfigImplementation", is_using_shared_caches: bool
    ) -> "DataConfigImplementation":
        with self.mutable_clone() as updated_copy:
            updated_copy.is_using_shared_caches = is_using_shared_caches
            return updated_copy

    def get_random_key(self) -> str:
        return self.random_key

    def get_dir_root(self) -> str:
        return self.dir_root

    def set_dir_root(self: "DataConfigImplementation", dir_root: str) -> "DataConfigImplementation":
        with self.mutable_clone() as updated_copy:
            updated_copy.dir_root = dir_root
            return updated_copy

    def get_output_file_name_format(self) -> str:
        return self.output_file_name_format

    def get_is_generating_paired_videos(self) -> bool:
        return self.is_generating_paired_videos

    def get_video_range(self) -> AbstractRange:
        return IntRange(self.video_min, self.video_max)

    def get_frame_range(self) -> AbstractRange:
        return IntRange(0, self.frame_max)

    def create_data_generation_request(
        self: "DataConfigImplementation", video_range: AbstractRange, random_key: str, dir_root: str
    ) -> "DataConfigImplementation":
        with self.mutable_clone() as new_config:
            new_config.video_min = video_range.min_ge
            new_config.video_max = video_range.max_lt
            new_config.random_key = random_key
            new_config.dir_root = dir_root
            return new_config

    # TODO: I wish data config were not expected to know which features would be generated...
    def get_features(self) -> Tuple[str, ...]:
        raise Exception("New data configs do not know about features")


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Vector2(Serializable):
    x: float
    y: float

    def __len__(self) -> int:
        return 2

    def __getitem__(self, item: int) -> float:
        if item == 0:
            return self.x
        if item == 1:
            return self.y
        raise IndexError(f"Vector2 only has 2 items, cannot access index {item}")


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class Vector3(Serializable):
    x: float
    y: float
    z: float

    def __len__(self) -> int:
        return 3

    def __getitem__(self, item: int) -> float:
        if item == 0:
            return self.x
        if item == 1:
            return self.y
        if item == 2:
            return self.z
        raise IndexError(f"Vector2 only has 3 items, cannot access index {item}")


AnyVec2 = Union[Vector2, float, int]
AnyVec3 = Union[Vector3, float, int]

T = TypeVar("T")
FloatT = TypeVar("FloatT", AnyVec2, AnyVec3, float)


class BaseDistribution(Serializable, Generic[T]):
    def sample(self, rand: Random) -> T:
        raise NotImplementedError()


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True, init=True)
class ChoicesDistribution(BaseDistribution[T]):
    value_choice: Tuple[T, ...]
    value_chance: Tuple[float, ...]

    @staticmethod
    def create_from_tuple(
        value_choice: Tuple[T, ...],
    ):
        return ChoicesDistribution(
            value_choice=tuple(value_choice),
            value_chance=(1.0,) * len(value_choice),
        )

    @staticmethod
    def create_from_dict(
        value_choice: Dict[T, float],
    ):
        return ChoicesDistribution(
            value_choice=tuple(value_choice),
            value_chance=tuple(value_choice.values()),
        )

    def sample(self, rand: Random) -> T:
        return rand.choices(self.value_choice, self.value_chance)[0]


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True, init=False)
class ConstDistribution(BaseDistribution[T]):
    value: T

    def __init__(self, value: T) -> None:
        self.value = value

    def sample(self, rand: Random) -> T:
        return self.value


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True, init=False)
class GaussianDistribution(BaseDistribution[FloatT]):
    value_mean: FloatT
    value_std: FloatT

    def __init__(self, value_mean: FloatT, value_std: FloatT) -> None:
        self.value_mean = value_mean
        self.value_std = value_std

    def sample(self, rand: Random) -> FloatT:
        if isinstance(self.value_mean, float):
            assert isinstance(self.value_std, float)
            return rand.gauss(self.value_mean, self.value_std)
        else:
            assert hasattr(self.value_mean, "__len__")
            assert not isinstance(self.value_std, float)
            values = []
            for i in range(0, len(self.value_mean)):
                value = rand.gauss(self.value_mean[i], self.value_std[i])
                values.append(value)
            return (self.value_mean.__class__)(*values)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True, init=False)
class BernoulliDistribution(BaseDistribution[bool]):
    probability: float
    value: bool

    def __init__(self, probability: float, value: bool = True) -> None:
        self.probability = probability
        self.value = value


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True, init=False)
class UniformDistribution(BaseDistribution[FloatT]):
    value_min: FloatT
    value_max: FloatT

    def __init__(self, value_min: FloatT, value_max: FloatT) -> None:
        self.value_min = value_min
        self.value_max = value_max

    def sample(self, rand: Random) -> FloatT:
        if isinstance(self.value_min, (bool, str)):
            raise Exception("Cannot use Uniform with bools or strings, that is nonsense")
        elif isinstance(self.value_min, (float, int)):
            assert isinstance(self.value_max, (float, int))
            value = (rand.random() * (self.value_max - self.value_min)) + self.value_min
            if isinstance(self.value_min, int):
                return round(value)
            return value
        else:
            assert hasattr(self.value_min, "__len__")
            assert not isinstance(self.value_max, float)
            values = []
            for i in range(0, len(self.value_min)):
                value = (rand.random() * (self.value_max[i] - self.value_min[i])) + self.value_min[i]
                values.append(value)
            return (self.value_min.__class__)(*values)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class FloatRange(IntRange):
    min_ge: float  # type: ignore
    max_lt: float  # type: ignore

    def overlap(self, other: "FloatRange") -> Optional["FloatRange"]:  # type: ignore
        min_ge = max(self.min_ge, other.min_ge)
        max_lt = min(self.max_lt, other.max_lt)
        if min_ge >= max_lt:
            return None
        return FloatRange(min_ge, max_lt)



import os
import re
import shutil
import sys
from pathlib import Path
from textwrap import dedent
from textwrap import indent
from typing import Dict
from typing import Final
from typing import Iterable
from typing import Optional
from typing import Sequence
from typing import Union
from urllib.request import urlretrieve
from zipfile import ZipFile

import fire

import avalon.install_godot_binary as binary
from avalon.common.utils import AVALON_PACKAGE_DIR
from avalon.contrib.utils import run_local_command
from avalon.datagen.generate_worlds import generate_evaluation_worlds
from avalon.datagen.godot_env.interactive_godot_process import AVALON_GODOT_PROJECT_PATH
from avalon.datagen.godot_env.interactive_godot_process import AVALON_HUMAN_WORLDS_PATH
from avalon.datagen.godot_env.interactive_godot_process import GODOT_EDITOR_PATH
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.constants import AvalonTaskGroup
from avalon.datagen.world_creation.constants import get_all_tasks_for_task_groups
from avalon.datagen.world_creation.world_generator import GenerateAvalonWorldParams
from avalon.datagen.world_creation.world_generator import generate_world
from avalon.datagen.world_creation.worlds.export import get_world_slug

# Must be created with `adb shell mkdir -p /sdcard/Android/data/com.godotengine.datagen`
MOCK_OCULUS_EXTERNAL_STORAGE: Final = "/sdcard/Android/data/com.godotengine.datagen/files"

GODOT_SOURCE_PATH: Final = Path(AVALON_GODOT_PROJECT_PATH)
GODOT_PROJECT_FILE_PATH: Final = f"{AVALON_GODOT_PROJECT_PATH}/project.godot"
OPENXR_PLUGIN_RELEASE: Final = "https://github.com/GodotVR/godot_openxr/releases/download/1.3.0/godot-openxr.zip"
OPENXR_PLUGIN_PATH: Final = GODOT_SOURCE_PATH / "addons/godot-openxr"


def _parse_task_groups(task_group_list: Union[str, Sequence[str]]) -> Iterable[AvalonTaskGroup]:
    mapping = {str(task.value).lower(): task for task in AvalonTaskGroup}
    if isinstance(task_group_list, str):
        task_group_list = task_group_list.split(",")
    for task_group in [t.strip().lower() for t in task_group_list]:
        assert task_group in mapping, (
            f"Invalid Avalon Task or Task Group {task_group}. "
            f"Allowed options (case insensitive): {', '.join(mapping.keys())}"
        )
        yield mapping[task_group]


def _rerun_this_command_with_env(env: Dict[str, str]) -> None:
    _, *args = sys.argv
    new_env = os.environ.copy()
    new_env.update(env)
    os.execve(sys.executable, [sys.executable, "-m", "avalon.for_humans", *args], new_env)


def _edit_project_settings(replacements: Dict[str, str]) -> None:
    with open(GODOT_PROJECT_FILE_PATH, "r+") as project_file:
        settings = project_file.read()
        for search_pattern, replacement_pattern in replacements.items():
            settings = re.sub(search_pattern, replacement_pattern, settings, flags=re.MULTILINE)
        project_file.seek(0)
        project_file.write(settings)
        project_file.truncate()


class VRHelperCLI:
    """CLI for human-centric usage of avalon (world generation, editor config, launching the editor)"""

    def available_tasks(self, exclude_groups: bool = False) -> None:
        """print available tasks and task groups for generation

        :param no_groups: Suppress task group output
        """
        task_group_lines = []
        single_tasks = []
        for task_group in AvalonTaskGroup:
            tasks = get_all_tasks_for_task_groups([task_group])
            if len(tasks) == 1 and tasks[0].name == task_group.name:
                single_tasks.append(tasks[0].name)
            else:
                task_group_lines.append(f"  {task_group.name}: f{', '.join(task.name for task in tasks)}")
        output_lines = [
            "Available Tasks:",
            f"  {', '.join(single_tasks)}\n",
        ]
        if not exclude_groups:
            output_lines.extend(["Available Task Groups:", *task_group_lines])
        print("\n".join(output_lines))

    def _ensure_hashseed_is_set(self):
        hashseed_env = "PYTHONHASHSEED"
        if os.environ.get(hashseed_env, None) is None:
            print(f"Rerunning command with {hashseed_env}=0 so level generation is deterministic\n")
            return _rerun_this_command_with_env({hashseed_env: "0"})

    def generate_evaluation_worlds(
        self,
        tasks: Union[str, Sequence[str]],
        for_practice: bool = False,
        start_seed: int = 10_000,
        worlds_per_task: int = 10,
        min_difficulty: float = 0.0,
        delete_existing_worlds: bool = False,
        verbose: bool = False,
    ) -> None:
        """Generate evaluation or practice worlds and save them to the internal `worlds` directory for inspection & playing by a human.

        Evaluation worlds will always ensure a special set of configurations are applied to guarantee a completeness of variety in certain tasks,
        where as `--for_practice` will use more randomness in it's configuration more akin to that seen in training.

        Note: to generate a single specific world identical to one seen in training, use `generate_single_world`.

        Examples:
            generate_evaluation_worlds easy,compositional --for_practice --delete_existing_worlds
            generate_evaluation_worlds --tasks="eat,move,navigate"
            generate_evaluation_worlds EAT,SURVIVE --worlds_per_task=3

        For a complete list of tasks see run `available_tasks`

        :param tasks: List of avalon tasks or task groups to generate (case insensitive, comma delimited).
        "param for_practice: Disable special evaluation configuration and output levels accessible by the left-hand "Practice World" teleporter.
        :param start_seed: Change the initial world generator random seed.
        :param worlds_per_task: Number of worlds to generate for each task.
        :param min_difficulty: Minimum difficulty with which to generate tasks.
        :param delete_existing_worlds: Delete all existing generated worlds before generating new ones.
        :param verbose: Enable verbose logging in world generator.
        """
        self._ensure_hashseed_is_set()

        avalon_tasks = get_all_tasks_for_task_groups(list(_parse_task_groups(tasks)))
        avalon_tasks = list(sorted(avalon_tasks, key=lambda g: g.name))

        if delete_existing_worlds:
            shutil.rmtree(AVALON_HUMAN_WORLDS_PATH, ignore_errors=True)

        generate_evaluation_worlds(
            is_recreating=True,
            is_practice=for_practice,
            is_generating_for_human=True,
            num_workers=4,
            base_output_path=Path(AVALON_HUMAN_WORLDS_PATH),
            min_difficulty=min_difficulty,
            tasks=avalon_tasks,
            num_worlds_per_task=worlds_per_task,
            start_seed=start_seed,
            is_verbose=verbose,
        )

    def generate_single_world(
        self,
        task: str,
        seed: int,
        index: int,
        difficulty: float,
        delete_existing_worlds: bool = False,
    ) -> None:
        """Generate a single world and save it to the internal `worlds` directory as a practice world for inspection & playing by a human.

        The output world will be identical to what would be seen in training, given the same configuration
        (Note that index effects the world generator).

        Examples:
            generate_single_worlds EAT --seed=2 --index=100 --difficulty=0.5

        For a complete list of tasks see run `available_tasks --exclude_groups`

        :param task: Task to generate (groups are invalid here)
        :param seed: World generator base random seed.
        :param index: Index of world generated during training (effects world generation).
        :param delete_existing_worlds: Delete all existing generated worlds before generating new ones.
        """
        self._ensure_hashseed_is_set()

        avalon_tasks = [t for t in AvalonTask if t.name.lower() == task.lower()]
        assert len(avalon_tasks) == 1, f"No such task {task}"
        avalon_task = avalon_tasks[0]

        if delete_existing_worlds:
            shutil.rmtree(AVALON_HUMAN_WORLDS_PATH, ignore_errors=True)

        slug = get_world_slug(avalon_task.name, seed=seed, index=index, difficulty=difficulty, is_practice=True)

        generate_world(
            GenerateAvalonWorldParams(
                output=f"{AVALON_HUMAN_WORLDS_PATH}/{slug}",
                task=avalon_task,
                difficulty=difficulty,
                seed=seed,
                index=index,
            )
        )

    def print_godot_project_path(self) -> None:
        print(GODOT_SOURCE_PATH)

    def setup_android_export_presets(
        self,
        keystore_path: str,
        keystore_user: str = "androiddebugkey",
        keystore_password: str = "android",
        export_apk_path: Optional[str] = None,
        debug_template_apk_path: Optional[str] = None,
        release_template_apk_path: Optional[str] = None,
        overwrite_apk_templates: bool = False,
    ) -> None:
        """Fill in the godot android export_presets.cfg template for running on oculus.

        All APK paths default to avalon/bin/{kind}.apk

        For more details see https://docs.godotengine.org/en/stable/tutorials/export/exporting_for_android.html

        :param keystore_path: Android debug.keystore path
        :param keystore_user: Android debug.keystore user
        :param keystore_password: Android debug.keystore password
        :param export_apk_path: Override export apk path. Default: avalon/bin/avalon.apk
        :param debug_template_apk_path: Override debug apk template path. Default: avalon/bin/debug_template.apk
        :param release_template_apk_path: Override release apk template path. Default: avalon/bin/release_template.apk
        :param overwrite_apk_templates: Overwrite pre-existing APK templates, if any exist
        """
        if export_apk_path is None:
            export_apk_path = f"{AVALON_PACKAGE_DIR}/bin/avalon.apk"
        if debug_template_apk_path is None:
            debug_template_apk_path = f"{AVALON_PACKAGE_DIR}/bin/debug_template.apk"
        if release_template_apk_path is None:
            release_template_apk_path = f"{AVALON_PACKAGE_DIR}/bin/release_template.apk"

        if not Path(debug_template_apk_path).exists() or overwrite_apk_templates:
            print("Downloading Debug APK template")
            urlretrieve("https://mmap.monster/godot/templates/android_debug.apk", debug_template_apk_path)

        if not Path(release_template_apk_path).exists() or overwrite_apk_templates:
            print("Downloading Release APK template")
            urlretrieve("https://mmap.monster/godot/templates/android_release.apk", release_template_apk_path)

        print("Configuring export_presets.cfg")
        with open(GODOT_SOURCE_PATH / "export_presets.template.cfg", "r") as template_file:
            template = template_file.read()
        replacements = {
            "/your/path/avalon.apk": export_apk_path,
            "/your/path/android_debug.apk": debug_template_apk_path,
            "/your/path/android_release.apk": release_template_apk_path,
            "/your/path/android.keystore": keystore_path,
            "yourdebuguser (maybe godot or androiddebugkey)": keystore_user,
            'keystore/debug_password="android"': f'keystore/debug_password="{keystore_password}"',
            'keystore/release_password="android"': f'keystore/release_password="{keystore_password}"',
        }
        for placeholder, value in replacements.items():
            template = template.replace(placeholder, value)

        with open(GODOT_SOURCE_PATH / "export_presets.cfg", "w") as preset_file:
            preset_file.write(template)
        print("Android export configured. You will need to restart the editor to see changes.")

    def remove_openxr_plugin(self) -> None:
        """Remove the openxr plugin installed with `install_openxr_plugin` and attempts to clean the project configuration.

        Closing the editor before running ths command is recommended.
        """
        shutil.rmtree(OPENXR_PLUGIN_PATH)
        _edit_project_settings(
            {
                "res://addons/godot-openxr/plugin.cfg": "",
                "res://addons/godot-openxr/config/godot_openxr.gdnlib": "",
            }
        )
        print("OpenXR plugin successfully removed and disabled.")

    def _configure_openxr_plugin(self) -> None:
        print(f"Configuring project settings...")
        _find_enabled_plugin_array = r"^(\[editor_plugins\](.|\n)*enabled=PoolStringArray\(.*) \)$"
        _append_openxr_plugin = r'\1"res://addons/godot-openxr/plugin.cfg" )'
        _find_gdnative_singleton_list = r"^(\[gdnative\](.|\n)*singletons=\[.*) \]$"
        _append_openxr_singleton = r'\1"res://addons/godot-openxr/config/godot_openxr.gdnlib" ]'
        _edit_project_settings(
            {
                _find_enabled_plugin_array: _append_openxr_plugin,
                _find_gdnative_singleton_list: _append_openxr_singleton,
            }
        )

    def install_openxr_plugin(self, overwrite: bool = False) -> None:
        """Installs and enables Godot's OpenXR plugin for VR support.

        Closing the editor before running ths command is recommended.

        The extension can be toggled in the editor under Project > Project Settings > Plugins,
        and can be removed with `remove_openxr_plugin`

        For more details see the godot docs at https://docs.godotengine.org/en/stable/tutorials/vr/openxr
        and our running in VR guide at https://github.com/Avalon-Benchmark/avalon/docs/running_in_vr.md#installing-openxr

        :param overwrite: Overwrite any existing installation
        """
        if OPENXR_PLUGIN_PATH.exists():
            if not overwrite:
                print(
                    "Cancelling: Already installed. Pass --overwrite if you want to overwrite the existing installation"
                )
                exit(0)
            shutil.rmtree(OPENXR_PLUGIN_PATH)

        print(f"Installing openxr plugin to {OPENXR_PLUGIN_PATH}")
        addon_path = OPENXR_PLUGIN_PATH.parent
        addon_path.mkdir(parents=False, exist_ok=True)
        archive_path = addon_path / "godot-openxr.zip"

        print(f"Downloading {OPENXR_PLUGIN_RELEASE} into addons dir...")
        urlretrieve(OPENXR_PLUGIN_RELEASE, archive_path)

        print(f"Extracting archive...")
        with ZipFile(archive_path, "r") as archive:
            archive.extractall(addon_path)

        print(f"Cleaning up...")
        os.remove(archive_path)
        shutil.move(addon_path / "godot_openxr_1.3.0/addons/godot-openxr", OPENXR_PLUGIN_PATH)
        shutil.rmtree(OPENXR_PLUGIN_PATH / "assets")
        shutil.rmtree(OPENXR_PLUGIN_PATH / "scenes")

        self._configure_openxr_plugin()

        print("OpenXR plugin successfully installed")

    def launch_editor(self, verbose: bool = False) -> None:
        """Launch the godot editor installed with `python -m avalon.install_godot_binary` pointing at the internal godot project

        .. caution::
           The included editor is primarily intended for inspection and debugging.
           Changing settings or editing files in the editor will effect avalon's runtime, and may degrade performance or otherwise break training.
           If you want to experiment with the code itself, prefer forking https://github.com/Avalon-Benchmark/avalon instead.

        :param verbose: Enable godot's verbose mode
        """
        if not os.path.exists(GODOT_EDITOR_PATH):
            assert binary.EDITOR in binary.available_builds(), "No editor available for this platform."
            install = "python -m avalon.install_godot_binary"
            print(f"Warning: no file found in GODOT_EDITOR_PATH. Attempting to run `{install}`")
            run_local_command(install)

        note = str(self.launch_editor.__doc__).split(".. caution::", 1)[1].split("\n\n")[0]
        print(f"\nCAUTION:{indent(dedent(note), '  ')}\n")

        run_local_command(f"{GODOT_EDITOR_PATH} --editor{' --verbose' if verbose else ''} {GODOT_PROJECT_FILE_PATH}")


if __name__ == "__main__":
    fire.Fire(VRHelperCLI)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import logging
import multiprocessing as mp
import time

import libtorchbeast
import numpy as np

# yapf: disable
from loguru import logger

from avalon.agent.torchbeast import atari_wrappers
from avalon.agent.torchbeast import avalon_helpers
from avalon.agent.torchbeast.avalon_helpers import create_godot_env
from avalon.agent.torchbeast.avalon_helpers import godot_config_from_flags
from avalon.common.log_utils import configure_parent_logging

parser = argparse.ArgumentParser(description='Remote Environment Server')

parser.add_argument("--pipes_basename", default="unix:/tmp/polybeast",
                    help="Basename for the pipes for inter-process communication. "
                    "Has to be of the type unix:/some/path.")
parser.add_argument('--num_servers', default=8, type=int, metavar='N',
                    help='Number of environment servers.')
parser.add_argument('--env', type=str, default='godot',
                    help='Gym environment.')
parser.add_argument('--env_seed', default=0, type=int,
                    help='Seed for environment.')
parser.add_argument('--task_difficulty_update', default=5e-3, type=float,
                    help='Update rate for env task difficulty.')
parser.add_argument('--meta_difficulty_update', default=2e-4, type=float,
                    help='Update rate for env meta difficulty.')
parser.add_argument('--energy_cost_coefficient', default=1e-8, type=float,
                    help='Coefficient for energy usage')
parser.add_argument('--head_pitch_coefficient', default=0, type=float,
                    help='Coefficient for body kinetic energy')
parser.add_argument('--head_roll_coefficient', default=0, type=float,
                    help='Coefficient for body kinetic energy')
parser.add_argument('--time_limit', default=120, type=int,
                    help='Time limit at 0 difficulty.')
parser.add_argument('--training_protocol', default="multi_task_basic",
                    help='Training protocol from paper')
parser.add_argument("--is_task_curriculum_disabled", action="store_true",
                    help="Is task curriculum disabled during training")
parser.add_argument("--is_meta_curriculum_enabled", action="store_true",
                    help="Is meta curriculum enabled during training")
parser.add_argument('--fixed_world_max_difficulty', default=0.5, type=float,
                    help='Fixed world max difficulty')
parser.add_argument('--fixed_world_path', type=str, default='',
                    help='Path for loading eval worlds')
parser.add_argument('--fixed_world_key', type=str, default='',
                    help='Key for loading eval worlds')
parser.add_argument('--load_checkpoint_filename', type=str, default='',
                    help='Checkpoint filename to load')

parser.add_argument("--energy_cost_aggregator", default="sum",
                    help="How to aggregate energy costs.")
# yapf: enable


class Env:
    def reset(self):
        logger.debug("reset called")
        return np.ones((4, 84, 84), dtype=np.uint8)

    def step(self, action):
        frame = np.zeros((4, 84, 84), dtype=np.uint8)
        return frame, 0.0, False, {}  # First three mandatory.


def create_env(env_name, config):
    if env_name == "godot":
        return create_godot_env(config)
    else:
        return atari_wrappers.wrap_pytorch(
            avalon_helpers.wrap_atari(
                atari_wrappers.wrap_deepmind(
                    atari_wrappers.make_atari(env_name),
                    clip_rewards=False,
                    frame_stack=True,
                    scale=False,
                )
            )
        )


def serve(env_name, server_address, config) -> None:
    configure_parent_logging()
    logging.info(f"serve called with config {config}")
    init = Env if env_name == "Mock" else lambda: create_env(env_name, config)
    server = libtorchbeast.Server(init, server_address=server_address)
    server.run()


def main(flags, exit_event: mp.Event):
    if not flags.pipes_basename.startswith("unix:"):
        raise Exception("--pipes_basename has to be of the form unix:/some/path.")

    processes = []
    for i in range(flags.num_servers):
        env_config = godot_config_from_flags(flags, env_index=i)
        p = mp.Process(target=serve, args=(flags.env, f"{flags.pipes_basename}.{i}", env_config))
        p.start()
        processes.append(p)

    try:
        # We are only here to listen to the interrupt.
        while True:
            time.sleep(1)
            if exit_event.is_set():
                logger.debug("exit event set")
                for p in processes:
                    p.kill()
                return
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    flags = parser.parse_args()
    main(flags)



from pathlib import Path
from typing import List
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

import attr
import numpy as np
from scipy import stats

from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.entities.altar import Altar
from avalon.datagen.world_creation.entities.constants import FOOD_TREE_BASE_HEIGHT
from avalon.datagen.world_creation.entities.food import CANONICAL_FOOD
from avalon.datagen.world_creation.entities.food import FOODS
from avalon.datagen.world_creation.entities.food import Food
from avalon.datagen.world_creation.entities.food import FoodTree
from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.entities.tools.weapons import Rock
from avalon.datagen.world_creation.entities.tools.weapons import Stick
from avalon.datagen.world_creation.entities.utils import get_random_ground_points
from avalon.datagen.world_creation.types import Point2DNP
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import get_rock_probability
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class EatTaskConfig(TaskConfig):
    # on the easiest difficulties, food starts out sitting on an altar which is right in front of you
    # this enables even random agents to succeed at the task and probably makes it much easier for agents to learn
    # this sparse reward setting.
    # these parameters control the size of the world in both modes (altar and non-altar, ie, normal)
    altar_extra_world_goal_dist_easy: float = 0.0
    altar_extra_world_goal_dist_hard: float = 5.0
    normal_extra_world_goal_dist_easy: float = 0.0
    normal_extra_world_goal_dist_hard: float = 10.0
    # parameters that control the size of the altar:
    altar_height_default: float = 1.3
    altar_height_variation_min: float = -1.0
    altar_height_variation_max: float = 0.5
    altar_width_default: float = 1.0
    altar_width_variation_min: float = 0.0
    altar_width_variation_max: float = 1.0
    # the probability that only a single food is spawned (vs multiple) at difficulty 0.0 and 1.0 respectively
    is_single_food_probability_easy: float = 1.0
    is_single_food_probability_hard: float = 0.3
    # how much food to create at the hardest difficulty. The agent must consume all of it to succed at the level
    max_food: int = 5
    # probability of whether food is on the ground or the tree at difficulty 0.0 and 1.0 respectively
    is_on_ground_probability_easy: float = 1.0
    is_on_ground_probability_hard: float = 0.1
    # probability of whether food is already opened at difficulty 0.0 and 1.0 respectively
    is_opened_probability_easy: float = 1.0
    is_opened_probability_hard: float = 0.05
    # how far away to spawn the foods beyond the first
    food_radius_min: float = 0.75
    food_radius_max: float = 2.0
    # at most how many extra tools to spawn
    extra_tool_count_max: int = 4


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class ForcedFood:
    food: Optional[Type[Food]] = None


def generate_eat_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: EatTaskConfig = EatTaskConfig(),
    _FORCED: Optional[ForcedFood] = None,
) -> None:
    world, locations, difficulty = create_eat_obstacle(
        rand, difficulty, export_config, task_config=task_config, _FORCED=_FORCED
    )
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)
    export_world(output_path, rand, world)


def create_eat_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    _FORCED: Optional[ForcedFood] = None,
    task_config: EatTaskConfig = EatTaskConfig(),
) -> Tuple[World, WorldLocations, float]:
    if _FORCED is not None and _FORCED.food:
        food_type = _FORCED.food()
    else:
        food_type, difficulty = select_categorical_difficulty(FOODS, difficulty, rand)
    is_single_food, difficulty = select_boolean_difficulty(
        difficulty,
        rand,
        initial_prob=task_config.is_single_food_probability_easy,
        final_prob=task_config.is_single_food_probability_hard,
    )
    # some things always come in bunches
    if food_type.is_always_multiple():
        is_single_food = False

    is_extra_tree_required = False
    on_object: Optional[Union[FoodTree, Altar]] = None
    if food_type.is_found_on_ground:
        is_on_altar, difficulty = select_boolean_difficulty(difficulty, rand)
    else:
        is_on_altar = False

    if is_on_altar:
        on_object = Altar.build(
            rand.uniform(task_config.altar_height_variation_min, task_config.altar_height_variation_max) * difficulty
            + task_config.altar_height_default,
            table_dim=task_config.altar_width_default
            + difficulty * rand.uniform(task_config.altar_width_variation_min, task_config.altar_width_variation_max),
        )
        desired_food_dist = (
            rand.uniform(task_config.altar_extra_world_goal_dist_easy, task_config.altar_extra_world_goal_dist_hard)
            * difficulty
            + 0.5
        )
    else:
        if food_type.is_grown_on_trees:
            is_on_ground = False
            if food_type.is_found_on_ground:
                is_on_ground, difficulty = select_boolean_difficulty(
                    difficulty,
                    rand,
                    initial_prob=task_config.is_on_ground_probability_easy,
                    final_prob=task_config.is_on_ground_probability_hard,
                )

            is_extra_tree_required = is_on_ground
            if not is_on_ground:
                tree_height = FOOD_TREE_BASE_HEIGHT
                on_object = FoodTree.build(
                    tree_height,
                    is_food_on_tree=True,
                )
        elif food_type.is_openable:
            is_opened, difficulty = select_boolean_difficulty(
                difficulty,
                rand,
                initial_prob=task_config.is_opened_probability_easy,
                final_prob=task_config.is_opened_probability_hard,
            )
            if is_opened:
                food_type = food_type.get_opened_version()

        desired_food_dist = (
            rand.uniform(task_config.normal_extra_world_goal_dist_easy, task_config.normal_extra_world_goal_dist_hard)
            * difficulty
            + 1
        )

    food_height = on_object.get_food_height(food_type) if on_object else food_type.get_offset()

    world, locations = create_world_from_constraint(
        stats.norm(desired_food_dist, desired_food_dist / 5.0),
        rand,
        difficulty,
        export_config,
        constraint,
        food_height,
        ideal_shore_dist_options=(0.5,),
    )
    world, locations = world.end_height_obstacles(locations, is_accessible_from_water=True)

    tree: Optional[FoodTree] = None
    if on_object and isinstance(on_object, Altar):
        actual_distance_to_food = np.linalg.norm(locations.goal - locations.spawn)
        # orient the altar towards you if you're really close
        if actual_distance_to_food < 1.6:
            direction_to_target = (locations.goal - locations.spawn) / actual_distance_to_food
            # noinspection PyTypeChecker
            yaw = np.angle(complex(direction_to_target[2], direction_to_target[0]), deg=True) + 180
            on_object = attr.evolve(on_object, yaw=yaw)
        world = world.add_item(
            on_object.place(locations.goal - FOOD_TREE_BASE_HEIGHT * UP_VECTOR),
            reset_height_offset=on_object.get_offset(),
        )
    elif on_object and isinstance(on_object, FoodTree):
        tree = on_object
    elif is_extra_tree_required:
        tree_height = FOOD_TREE_BASE_HEIGHT
        tree = FoodTree.build(tree_height, is_food_on_tree=False)

    primary_food = food_type
    # add the tree if required, because we need to use it to figure out the spawn location of the other foods
    if tree is not None:
        # place the tree + optional base in case it would be floating.
        tree, tree_base = tree.place(locations.spawn, locations.goal, world.get_height_at)
        world = world.add_item(tree)
        tree = cast(FoodTree, world.items[-1])
        if tree_base is not None:
            world = world.add_item(tree_base)
        # attach the food + set on_object for future calculations
        if tree.is_food_on_tree:
            primary_food = primary_food.attached_to(tree)
            on_object = tree

    # add the foods
    food_count = 1
    if not is_single_food:
        max_food_delta = task_config.max_food - 2
        food_count = 2 + round(difficulty * max_food_delta * rand.uniform())

    # figure out where to position the foods
    food_locations = np.array([locations.goal])
    food_hover = food_type.get_offset()
    other_food_count = food_count - 1
    if other_food_count > 0:
        if on_object is None:
            if tree is None:
                other_food_locations = get_random_ground_points(
                    rand,
                    locations.goal,
                    other_food_count,
                    world.map,
                    task_config.food_radius_min,
                    task_config.food_radius_max,
                    food_hover,
                    locations.island,
                )
            else:
                other_food_locations = tree.get_food_locations(
                    rand,
                    locations.goal,
                    other_food_count,
                    world.map,
                    task_config.food_radius_min,
                    task_config.food_radius_max,
                    food_hover,
                    locations.island,
                )
        else:
            other_food_locations = on_object.get_food_locations(
                rand,
                locations.goal,
                other_food_count,
                world.map,
                task_config.food_radius_min,
                task_config.food_radius_max,
                food_hover,
                locations.island,
            )
        food_locations = np.concatenate([food_locations, other_food_locations], axis=0)
    for i, food_location in enumerate(food_locations):
        base = primary_food if i == 0 else food_type
        world = world.add_item(attr.evolve(base, position=base.additional_offset + food_location))
    # figure out what tools we need
    tools: List[Tool] = []
    first_tool = select_tool(food_type, rand, difficulty)
    if first_tool:
        tools.append(first_tool)
    # some chance of including some rocks / sticks naturally anyway
    if on_object and on_object.is_tool_helpful:
        if len(tools) > 0:
            # default to more tools if we need one
            extra_tool_count = round(
                task_config.extra_tool_count_max
                - (task_config.extra_tool_count_max - 1) * difficulty
                - rand.uniform() * difficulty
            )
        else:
            # default to fewer tools if we don't. Don't want to confuse and distract the agent on easier difficulties
            extra_tool_count = round(rand.uniform(0, 3) * difficulty)
        for _ in range(extra_tool_count):
            if rand.uniform() < get_rock_probability(difficulty):
                tools.append(Rock())
            else:
                tools.append(Stick())
    # position and add any tools
    tool_count = len(tools)
    if tool_count > 0:
        min_radius = 1.0
        max_radius = min_radius + 1.0 + difficulty * 3
        tool_offsets = np.array([x.get_offset() for x in tools])
        tool_locations = get_random_ground_points(
            rand,
            locations.spawn,
            tool_count,
            world.map,
            min_radius,
            max_radius,
            tool_offsets,
            locations.island,
        )
        for tool, tool_location in zip(tools, tool_locations):
            world = world.add_item(attr.evolve(tool, position=tool_location))

    return world, locations, difficulty


def add_food_tree(
    rand: np.random.Generator,
    difficulty: float,
    world: World,
    point: Point2DNP,
    is_tool_food_allowed: bool,
    spawn_point: Optional[Point2DNP] = None,
) -> Tuple[float, Optional[Tool], World]:
    # figure out what foods we're allowed to create
    allowed_foods = [x for x in FOODS if x.is_grown_on_trees]
    if not is_tool_food_allowed:
        allowed_foods = [x for x in allowed_foods if not x.is_tool_required]

    # figure out the exact food and what that entails
    food, difficulty = select_categorical_difficulty(allowed_foods, difficulty, rand)
    is_on_ground = False
    if food.is_found_on_ground:
        is_on_ground, difficulty = select_boolean_difficulty(difficulty, rand, final_prob=0.1)

    tree = FoodTree.build(
        tree_height=FOOD_TREE_BASE_HEIGHT,
        is_food_on_tree=not is_on_ground,
    )
    tool: Optional[Tool] = None
    if is_tool_food_allowed:
        tool = select_tool(food, rand, difficulty)
    if is_on_ground:
        is_opened, difficulty = select_boolean_difficulty(difficulty, rand, final_prob=0.05)
        if is_opened:
            food = food.get_opened_version()

    # figure out positioning and orientation
    food_height = tree.get_food_height(food)
    food_location = np.array([point[0], world.get_height_at(point) + food_height, point[1]])
    if spawn_point is None:
        random_vec = np.array([rand.uniform(), rand.uniform(), rand.uniform(0.001)])
        spawn_location = food_location + random_vec
    else:
        spawn_location = np.array([spawn_point[0], 0, spawn_point[1]])
    food = attr.evolve(food, position=food_location)

    world = add_food_and_tree(food, spawn_location, tree, world)

    return difficulty, tool, world


def add_food_and_tree(
    food: Food, spawn_location: Point3DNP, tree: FoodTree, world: World, is_base_allowed: bool = True
) -> World:
    # put the tree (and any required base) into the world
    tree, tree_base = tree.place(spawn_location, food.position, world.get_height_at)
    world = world.add_item(tree)
    tree = cast(FoodTree, world.items[-1])
    if tree_base is not None and is_base_allowed:
        world = world.add_item(tree_base)
    # attach the food + set on_object for future calculations
    if tree.is_food_on_tree:
        food = food.attached_to(tree)
    # add the food to the world
    return world.add_item(food)


def add_food_tree_for_simple_task(world: World, locations: WorldLocations) -> World:
    food = attr.evolve(CANONICAL_FOOD, position=locations.goal)
    tree = FoodTree.build(
        tree_height=FOOD_TREE_BASE_HEIGHT,
        is_food_on_tree=True,
    )
    return add_food_and_tree(food, locations.spawn, tree, world, is_base_allowed=False)


def select_tool(food: Food, rand: np.random.Generator, difficulty: float) -> Optional[Tool]:
    odds, options = food.get_tool_options()
    if odds == 0:
        return None
    if odds == 1.0 or difficulty_variation(0.0, 1.0, rand, difficulty) <= odds:
        if len(options) == 1:
            return options[0]()
        return rand.choice(options)()  # type: ignore
    return None



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import Final
from typing import Generic
from typing import Iterable
from typing import List
from typing import Optional
from typing import Protocol
from typing import Set
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

import attr
import numpy as np
from gym import spaces
from numpy import typing as npt

from avalon.common.type_utils import assert_not_none
from avalon.common.utils import only
from avalon.datagen.godot_generated_types import FAKE_TYPE_IMAGE as GENERATED_FAKE_TYPE_IMAGE

# Mapping of feature name to (data_type, shape).
FeatureSpecDict = OrderedDict[str, Tuple[int, Tuple[int, ...]]]

# Data dictionary of features, interpretable by a `FeatureSpecDict`
FeatureDataDict = Dict[str, npt.NDArray]


OPTIONAL_FEATURES = ["isometric_rgbd", "top_down_rgbd"]

FAKE_TYPE_IMAGE: Final = GENERATED_FAKE_TYPE_IMAGE
TYPE_INT: Final = 2
TYPE_REAL: Final = 3
TYPE_VECTOR2: Final = 5
TYPE_VECTOR3: Final = 7
TYPE_QUAT: Final = 10

NP_DTYPE_MAP: Final[Dict[int, Type[np.number]]] = {
    TYPE_INT: np.int32,
    TYPE_REAL: np.float32,
    TYPE_VECTOR2: np.float32,
    TYPE_VECTOR3: np.float32,
    TYPE_QUAT: np.float32,
    FAKE_TYPE_IMAGE: np.uint8,
}


class InvalidObservationType(Exception):
    pass


class ObservationProtocol(Protocol):
    @classmethod
    def get_space_for_attribute(cls, feature: str, data_type: int, shape: Tuple[int, ...]) -> Optional[spaces.Space]:
        ...

    @classmethod
    def get_exposed_features(cls) -> Iterable[str]:
        """Features that should be exposed to the agent"""

    @classmethod
    def get_selected_features(cls) -> Iterable[str]:
        """Features to select and return from godot every step"""


ObservationType = TypeVar("ObservationType", bound=ObservationProtocol)


class AttrsObservation(ObservationProtocol):
    @classmethod
    def get_selected_features(cls) -> Tuple[str, ...]:
        return tuple(field.name for field in attr.fields(cls))


class GodotObservationContext(Generic[ObservationType]):
    def __init__(
        self,
        observation_type: Type[ObservationType],
        is_space_flattened: bool,
        available_features: FeatureSpecDict,
    ) -> None:
        self.observation_type = observation_type
        self.is_space_flattened = is_space_flattened
        self.available_features = available_features
        self.selected_features = self._select_features()
        self.observation_space: Union[spaces.Dict, spaces.Space] = self._create_observation_space()
        self.flattened_observation_keys: List[str] = []
        if is_space_flattened:
            (
                self.observation_space,
                self.flattened_observation_keys,
            ) = self._flatten_observation_space(self.observation_space)

    def _select_features(self) -> FeatureSpecDict:
        # validate our ObservationType and get selected_features
        selected_features: FeatureSpecDict = OrderedDict()
        for field in self.observation_type.get_selected_features():
            type_and_dims = self.available_features.get(field, None)
            # TODO: think of a cleaner way to make these optional?
            if type_and_dims is None and field in OPTIONAL_FEATURES:
                continue
            elif type_and_dims is None and field not in OPTIONAL_FEATURES:
                raise InvalidObservationType(
                    f"Could not find requested feature '{field}'. Available features are: {list(self.available_features)}"
                )
            # TODO: check that the types line up
            assert type_and_dims is not None
            selected_features[field] = type_and_dims
        return selected_features

    def _create_observation_space(self) -> spaces.Dict:
        # create our observation space
        observation_space_dict = {}
        exposed_features = self.observation_type.get_exposed_features()
        for feature_name, (data_type, dims) in self.available_features.items():
            if feature_name in exposed_features:
                if feature_name not in self.selected_features:
                    raise InvalidObservationType(f"Cannot expose feature {feature_name}!")
                space = self.observation_type.get_space_for_attribute(feature_name, data_type, dims)
                if space is None:
                    space = _get_default_space(data_type, dims)
                observation_space_dict[feature_name] = space

        return spaces.Dict(observation_space_dict)

    def _flatten_observation_space(self, observation_space: spaces.Dict):
        flattened_keys = sorted(list(observation_space.spaces.keys()))
        return flatten_observation_space(observation_space, flattened_keys), flattened_keys

    def _flatten_observation(self, observation: ObservationType) -> np.ndarray:
        return flatten_observation(observation, self.flattened_observation_keys)

    def lamify(self, observation: ObservationType):
        """Convert a well-typed observation to an Env-compliant observation space dict."""
        if self.is_space_flattened:
            return self._flatten_observation(observation)

        assert isinstance(self.observation_space, spaces.Dict)
        return {x: getattr(observation, x) for x in self.observation_space.spaces.keys()}

    def make_observation(self, feature_data: Dict[str, Any]) -> ObservationType:
        return self.observation_type(**feature_data)


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class AvalonObservation(AttrsObservation):
    rgbd: npt.NDArray[np.uint8]
    episode_id: npt.NDArray[np.int32]
    frame_id: npt.NDArray[np.int32]
    reward: npt.NDArray[np.float32]
    is_done: npt.NDArray[np.float32]
    is_dead: npt.NDArray[np.float32]

    target_head_position: npt.NDArray[np.float32]
    target_left_hand_position: npt.NDArray[np.float32]
    target_right_hand_position: npt.NDArray[np.float32]

    target_head_rotation: npt.NDArray[np.float32]
    target_left_hand_rotation: npt.NDArray[np.float32]
    target_right_hand_rotation: npt.NDArray[np.float32]

    physical_body_position: npt.NDArray[np.float32]
    physical_head_position: npt.NDArray[np.float32]
    physical_left_hand_position: npt.NDArray[np.float32]
    physical_right_hand_position: npt.NDArray[np.float32]
    physical_body_rotation: npt.NDArray[np.float32]
    physical_head_rotation: npt.NDArray[np.float32]
    physical_left_hand_rotation: npt.NDArray[np.float32]
    physical_right_hand_rotation: npt.NDArray[np.float32]
    physical_body_delta_position: npt.NDArray[np.float32]
    physical_head_delta_position: npt.NDArray[np.float32]
    physical_left_hand_delta_position: npt.NDArray[np.float32]
    physical_right_hand_delta_position: npt.NDArray[np.float32]
    physical_body_delta_rotation: npt.NDArray[np.float32]
    physical_head_delta_rotation: npt.NDArray[np.float32]
    physical_left_hand_delta_rotation: npt.NDArray[np.float32]
    physical_right_hand_delta_rotation: npt.NDArray[np.float32]
    physical_head_relative_position: npt.NDArray[np.float32]
    physical_left_hand_relative_position: npt.NDArray[np.float32]
    physical_right_hand_relative_position: npt.NDArray[np.float32]
    physical_head_relative_rotation: npt.NDArray[np.float32]
    physical_left_hand_relative_rotation: npt.NDArray[np.float32]
    physical_right_hand_relative_rotation: npt.NDArray[np.float32]
    left_hand_thing_colliding_with_hand: npt.NDArray[np.float32]
    left_hand_held_thing: npt.NDArray[np.float32]
    right_hand_thing_colliding_with_hand: npt.NDArray[np.float32]
    right_hand_held_thing: npt.NDArray[np.float32]

    nearest_food_position: npt.NDArray[np.float32]
    nearest_food_id: npt.NDArray[np.float32]
    is_food_present_in_world: npt.NDArray[np.float32]

    physical_body_kinetic_energy_expenditure: npt.NDArray[np.float32]
    physical_body_potential_energy_expenditure: npt.NDArray[np.float32]
    physical_head_potential_energy_expenditure: npt.NDArray[np.float32]
    physical_left_hand_kinetic_energy_expenditure: npt.NDArray[np.float32]
    physical_left_hand_potential_energy_expenditure: npt.NDArray[np.float32]
    physical_right_hand_kinetic_energy_expenditure: npt.NDArray[np.float32]
    physical_right_hand_potential_energy_expenditure: npt.NDArray[np.float32]
    fall_damage: npt.NDArray[np.float32]
    hit_points_lost_from_enemies: npt.NDArray[np.float32]
    hit_points_gained_from_eating: npt.NDArray[np.float32]
    hit_points: npt.NDArray[np.float32]

    # TODO: think of a cleaner way to make these optional?
    isometric_rgbd: npt.NDArray[np.uint8] = np.array([], dtype=np.uint8)
    top_down_rgbd: npt.NDArray[np.uint8] = np.array([], dtype=np.uint8)

    @classmethod
    def get_space_for_attribute(cls, feature: str, data_type: int, shape: Tuple[int, ...]) -> Optional[spaces.Space]:
        return None

    @classmethod
    def get_exposed_features(cls) -> Tuple[str, ...]:
        return (
            "rgbd",
            "physical_body_delta_position",
            "physical_body_delta_rotation",
            "physical_head_delta_position",
            "physical_left_hand_delta_position",
            "physical_right_hand_delta_position",
            "physical_head_delta_rotation",
            "physical_left_hand_delta_rotation",
            "physical_right_hand_delta_rotation",
            "physical_head_relative_position",
            "physical_left_hand_relative_position",
            "physical_right_hand_relative_position",
            "physical_head_relative_rotation",
            "physical_left_hand_relative_rotation",
            "physical_right_hand_relative_rotation",
            "left_hand_thing_colliding_with_hand",
            "left_hand_held_thing",
            "right_hand_thing_colliding_with_hand",
            "right_hand_held_thing",
            "fall_damage",
            "hit_points_lost_from_enemies",
            "hit_points_gained_from_eating",
            "hit_points",
            "frame_id",  # TODO: this will actually be frames_remaining
        )


def _get_default_space(data_type: int, dims: Tuple[int, ...]) -> spaces.Space:
    np_dtype = NP_DTYPE_MAP[data_type]
    if data_type == FAKE_TYPE_IMAGE:
        return spaces.Box(low=0, high=255, shape=dims, dtype=np_dtype)
    return spaces.Box(low=-np.inf, high=np.inf, shape=dims, dtype=np_dtype)


def flatten_observation_space(observation_space: spaces.Dict, flattened_observation_keys: List[str]) -> spaces.Space:
    lows, highs = [], []
    dtypes: Set[np.dtype] = set()
    for key in flattened_observation_keys:
        space = observation_space[key]
        assert isinstance(space, spaces.Box)
        lows.append(np.zeros(assert_not_none(space.shape), dtype=space.dtype) + space.low)
        highs.append(np.zeros(assert_not_none(space.shape), dtype=space.dtype) + space.high)
        dtypes.add(assert_not_none(space.dtype))

    dtype = only(dtypes)
    flat_lows = np.concatenate(lows, axis=-1)
    flat_highs = np.concatenate(highs, axis=-1)
    shape = flat_lows.shape
    return spaces.Box(low=flat_lows, high=flat_highs, shape=shape, dtype=assert_not_none(dtype))  # type: ignore[arg-type]


def flatten_observation(observation: ObservationType, flattened_observation_keys: List[str]) -> np.ndarray:
    return np.concatenate([getattr(observation, x) for x in flattened_observation_keys], axis=-1)



import math
from itertools import product
from pathlib import Path
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

import attr
import numpy as np
from scipy import stats

from avalon.common.errors import SwitchError
from avalon.common.utils import only
from avalon.datagen.errors import ImpossibleWorldError
from avalon.datagen.world_creation.configs.building import BuildingConfig
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import IndoorTaskConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.constants import BOX_HEIGHT
from avalon.datagen.world_creation.constants import FOOD_HOVER_DIST
from avalon.datagen.world_creation.constants import ITEM_FLATTEN_RADIUS
from avalon.datagen.world_creation.constants import JUMPING_REQUIRED_HEIGHT
from avalon.datagen.world_creation.constants import MAX_JUMP_HEIGHT_METERS
from avalon.datagen.world_creation.entities.item import Item
from avalon.datagen.world_creation.entities.tools.placeholder import Placeholder
from avalon.datagen.world_creation.entities.tools.stone import Stone
from avalon.datagen.world_creation.entities.tools.tool import Tool
from avalon.datagen.world_creation.geometry import BuildingTile
from avalon.datagen.world_creation.geometry import local_to_global_coords
from avalon.datagen.world_creation.indoor.builders import DefaultHallwayBuilder
from avalon.datagen.world_creation.indoor.builders import HouseLikeRoomBuilder
from avalon.datagen.world_creation.indoor.builders import RectangleFootprintBuilder
from avalon.datagen.world_creation.indoor.builders import WindowBuilder
from avalon.datagen.world_creation.indoor.building import Building
from avalon.datagen.world_creation.indoor.building import BuildingAestheticsConfig
from avalon.datagen.world_creation.indoor.building import BuildingTask
from avalon.datagen.world_creation.indoor.constants import DEFAULT_FLOOR_THICKNESS
from avalon.datagen.world_creation.indoor.constants import Azimuth
from avalon.datagen.world_creation.indoor.constants import TileIdentity
from avalon.datagen.world_creation.indoor.task_generator import CANONICAL_BUILDING_LOCATION
from avalon.datagen.world_creation.indoor.task_generator import BuildingTaskGenerator
from avalon.datagen.world_creation.indoor.task_generator import IndoorTaskParams
from avalon.datagen.world_creation.indoor.task_generator import _decide_boulder_mass
from avalon.datagen.world_creation.indoor.task_generator import add_food_island
from avalon.datagen.world_creation.indoor.task_generator import create_building_for_skill_scenario
from avalon.datagen.world_creation.indoor.task_generator import decide_spawn_room_and_tile
from avalon.datagen.world_creation.indoor.task_generator import make_indoor_task_world
from avalon.datagen.world_creation.indoor.task_generator import rectangle_dimensions_within_radius
from avalon.datagen.world_creation.indoor.tiles import decide_tiles_by_distance
from avalon.datagen.world_creation.indoor.utils import inset_borders
from avalon.datagen.world_creation.tasks.climb import replace_placeholder_with_goal
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.types import Point3DNP
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_categorical_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.utils import add_offsets
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class StackIndoorTaskConfig(IndoorTaskConfig):
    # Smallest possible site radius for a building with this task to be placed on (largest size is inherited)
    min_site_radius: float = 8.0
    # Standard deviation for the distribution deciding the site radius: higher means more variability at same difficulty
    site_radius_std_dev: float = 2.5
    # Total number of stories for the building. Note that this task is a single-story task, all other stories will
    # be purely aesthetic (e.g. to help with being viewed in the distance outdoors).
    story_count: int = 2
    # Min/max number of stones that need to be vertically stacked in order to solve the task
    min_stackable_height: int = 1
    max_stackable_height: int = 4
    # The task will create enough crates to stack a staircase + this many surplus stones (in case some get lost / stuck)
    surplus_stone_count: int = 0
    # Square island side length; should be large enough to prevent knocking food off by reaching or jumping
    food_island_size: int = 5
    # Min/max stone masses used to determine exact stone size; lighter stones are less stable
    min_stone_mass: float = 50
    max_stone_mass: float = 500


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class StackOutdoorTaskConfig(TaskConfig):
    # how much to move the boxes from their "canonical" (ie, solved) configuration at difficulty 0.0 and 1.0
    randomization_dist_easy: float = 0.0
    randomization_dist_hard: float = 6.0
    # makes the cliff at least this tall (with zero boxes). Final height will be num_boxes * box_height taller than this
    base_height: float = JUMPING_REQUIRED_HEIGHT - 0.5
    # how wide to make the region of the cliff where  the height is precisely correct for stacking. Other areas could
    # be taller, so making this value smaller often makes the task harder
    cliff_width_easy: float = 10.0
    cliff_width_hard: float = 1.0
    cliff_width_std_dev: float = 1.0
    # controls how much space will be present for stacking boxes at the base of the cliff
    cliff_space: float = 3.0


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class StackTaskConfig(TaskConfig):
    # likelihood that the indoor variant of the task will be used
    indoor_probability: float = 0.2
    # the configs for each variant of the task
    indoor_config: StackIndoorTaskConfig = StackIndoorTaskConfig()
    outdoor_config: StackOutdoorTaskConfig = StackOutdoorTaskConfig()


def generate_stack_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: StackTaskConfig = StackTaskConfig(),
) -> None:
    is_indoor = rand.uniform() < task_config.indoor_probability
    if is_indoor:
        building, entities, spawn_location, target_location = create_building_for_skill_scenario(
            rand,
            difficulty,
            StackTaskGenerator(task_config.indoor_config),
            position=CANONICAL_BUILDING_LOCATION,
            is_indoor_only=True,
        )
        world = make_indoor_task_world(
            building, entities, difficulty, spawn_location, target_location, rand, export_config
        )
    else:
        world, locations, difficulty = create_stack_obstacle(
            rand, difficulty, export_config, task_config=task_config.outdoor_config
        )
        world, locations = world.end_height_obstacles(
            locations, is_accessible_from_water=False, is_spawn_region_climbable=False
        )
        world = add_food_tree_for_simple_task(world, locations)
        world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal)

    export_world(output_path, rand, world)


def create_stack_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    is_for_carry: bool = False,
    task_config: StackOutdoorTaskConfig = StackOutdoorTaskConfig(),
) -> Tuple[World, WorldLocations, float]:

    stack_height = round(normal_distrib_range(0.51, 3.49, 0.25, rand, difficulty))
    spare_boxes, difficulty = select_categorical_difficulty([3, 2, 1, 0], difficulty, rand)

    desired_height = stack_height * BOX_HEIGHT + task_config.base_height

    # how far to jiggle the boxes from their "solved" configuration
    randomization_dist = scale_with_difficulty(
        difficulty, task_config.randomization_dist_easy, task_config.randomization_dist_hard
    )

    # you aren't technically going to have to move this far--the food gets updated to be close to the cliff edge
    # this mostly just gives a bit of space for the climb paths
    base_dist = 2.0 * desired_height + 2.0 * randomization_dist
    desired_goal_dist = difficulty_variation(base_dist, 2.0 * base_dist, rand, difficulty)

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, 0.5), rand, difficulty, export_config, constraint
    )

    # this ensures there is space at the base of the cliff
    if 2 * randomization_dist > locations.get_2d_spawn_goal_distance() - task_config.cliff_space:
        randomization_dist = (locations.get_2d_spawn_goal_distance() - task_config.cliff_space) / 2.0
    extra_safety_radius = randomization_dist

    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        gap_distance=0.0,
        height=desired_height,
        constraint=constraint,
        traversal_width=normal_distrib_range(
            task_config.cliff_width_easy,
            task_config.cliff_width_hard,
            task_config.cliff_width_std_dev,
            rand,
            difficulty,
        ),
        inner_traversal_length=0.0,
        is_single_obstacle=True,
        is_inside_climbable=False,
        inner_solution=HeightSolution(
            inside_items=_get_boxes(stack_height, spare_boxes),
            inside_item_randomization_distance=randomization_dist,
            inside_item_radius=0.5,
            solution_point_brink_distance=2.0,
            outside_items=tuple([Placeholder()]),
            outside_item_radius=1.0,
        ),
        extra_safety_radius=extra_safety_radius,
        probability_of_centering_on_spawn=0.0 if is_for_carry else 0.5,
    )
    ring_config = attr.evolve(ring_config, terrain_blurs=((8.0, 0.5), (6.0, 0.75), (4.0, 1.0)))
    earlier_item_count = len(world.items)
    world = world.add_height_obstacle(rand, ring_config, locations.island)
    world = flatten_places_under_items(world, earlier_item_count, Stone)

    # move the food to the edge so that it is visible
    new_locations, world = replace_placeholder_with_goal(locations, world)

    return world, new_locations, difficulty


def flatten_places_under_items(
    world: World, earlier_item_count: int, filter_to_classes: Union[Type[Item], Tuple[Type[Item], ...]]
):
    new_stones = []
    for i in range(earlier_item_count, len(world.items)):
        item = world.items[i]
        if isinstance(item, filter_to_classes):
            new_stones.append(item)
    stone_positions = [to_2d_point(x.position) for x in new_stones]
    for i, stone in enumerate(new_stones):
        distances = []
        for j, other_stone in enumerate(new_stones):
            if i == j:
                continue
            distances.append(np.linalg.norm(stone_positions[i] - stone_positions[j]))
        if len(distances) > 0 and np.min(distances) > ITEM_FLATTEN_RADIUS:
            world = world.flatten(stone_positions[i], ITEM_FLATTEN_RADIUS, ITEM_FLATTEN_RADIUS)
    return world


def _get_boxes(stack_height: int, spare_boxes: int) -> Tuple[Tool, ...]:
    if stack_height == 1:
        boxes = [
            Stone(position=np.array([0.0, 0.0, 0.0])),
        ]
    elif stack_height == 2:
        boxes = [
            Stone(position=np.array([0.0, 0.0, 0.0])),
            Stone(position=np.array([0.0, 1.0, 0.0])),
            Stone(position=np.array([-1.0, 0.0, 0.0])),
        ]
    elif stack_height == 3:
        boxes = [
            Stone(position=np.array([0.0, 0.0, 0.0])),
            Stone(position=np.array([0.0, 1.0, 0.0])),
            Stone(position=np.array([0.0, 2.0, 0.0])),
            Stone(position=np.array([-1.0, 0.0, 0.0])),
            Stone(position=np.array([-1.0, 1.0, 0.0])),
            Stone(position=np.array([-2.0, 0.0, 0.0])),
        ]
    else:
        raise SwitchError(f"Unhandled number of boxes: {stack_height}")

    if spare_boxes >= 1:
        boxes.append(Stone(position=np.array([0.0, 0.0, -1.0])))
    if spare_boxes >= 2:
        boxes.append(Stone(position=np.array([0.0, 0.0, 1.0])))
    if spare_boxes >= 3:
        boxes.append(Stone(position=np.array([-1.0, 0.0, 1.0])))
    assert spare_boxes <= 3

    return tuple(add_offsets(boxes))


@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)
class StackTaskGenerator(BuildingTaskGenerator):
    config: StackIndoorTaskConfig = StackIndoorTaskConfig()

    def get_site_radius(self, rand: np.random.Generator, difficulty: float) -> float:
        return normal_distrib_range(
            self.config.min_site_radius, self.config.max_site_radius, self.config.site_radius_std_dev, rand, difficulty
        )

    def get_building_config(
        self,
        rand: np.random.Generator,
        difficulty: float,
        radius: float,
        allowed_auxiliary_tasks: Tuple[BuildingTask, ...] = tuple(),
        aesthetics: BuildingAestheticsConfig = BuildingAestheticsConfig(),
    ) -> BuildingConfig:
        width, length = rectangle_dimensions_within_radius(radius)
        return BuildingConfig(
            width=width,
            length=length,
            story_count=self.config.story_count,
            footprint_builder=RectangleFootprintBuilder(),
            room_builder=HouseLikeRoomBuilder(max_rooms=1),
            story_linker=None,
            hallway_builder=DefaultHallwayBuilder(),
            window_builder=WindowBuilder(),
            aesthetics=aesthetics,
        )

    def get_principal_obstacle_params(self, rand: np.random.Generator, difficulty: float, building: Building) -> Tuple:
        first_story = building.stories[0]
        room = only(first_story.rooms)
        border_value = 0
        permitted_tiles = inset_borders(room.tiles, void_value=TileIdentity.ROOM.value, border_value=border_value)
        tile_zs, tile_xs = np.where(permitted_tiles == border_value)
        permitted_tile_tuples = list(zip(*(tile_zs, tile_xs)))
        spawn_room, spawn_tile = decide_spawn_room_and_tile(rand, first_story, {room.id: permitted_tile_tuples})

        min_stackable_height = self.config.min_stackable_height
        max_stackable_height = self.config.max_stackable_height
        stackable_height = round(min_stackable_height + difficulty * (max_stackable_height - min_stackable_height))
        required_stone_count = (stackable_height * (stackable_height + 1)) // 2
        surplus_stone_count = self.config.surplus_stone_count

        food_island_size = self.config.food_island_size
        food_island_height = (stackable_height * BOX_HEIGHT) + MAX_JUMP_HEIGHT_METERS * 0.95

        min_room_size = math.ceil(BOX_HEIGHT) * 2 + food_island_size
        if spawn_room.width < min_room_size or spawn_room.length < min_room_size:
            raise ImpossibleWorldError("Room too small for stack task")

        viable_food_island_tiles = list(
            product(
                range(1, spawn_room.width - food_island_size - 2), range(1, spawn_room.length - food_island_size - 2)
            )
        )
        if len(viable_food_island_tiles) == 0:
            raise ImpossibleWorldError("Building too small for push task")
        food_island_top_left_tile = only(
            decide_tiles_by_distance(viable_food_island_tiles, spawn_tile, difficulty, rand)
        )
        food_island_position = BuildingTile(*food_island_top_left_tile)
        food_island_tiles = list(
            product(
                range(food_island_position.x, food_island_position.x + food_island_size),
                range(food_island_position.z, food_island_position.z + food_island_size),
            )
        )

        free_tiles = [
            tile
            for tile in list(product(range(spawn_room.width), range(spawn_room.length)))
            if tile not in food_island_tiles and tile != spawn_tile
        ]
        food_island_center_tile = (
            food_island_top_left_tile[0] + food_island_size // 2,
            food_island_top_left_tile[1] + food_island_size // 2,
        )
        stone_tiles = decide_tiles_by_distance(
            free_tiles,
            food_island_center_tile,
            difficulty,
            rand,
            tile_count=required_stone_count + surplus_stone_count,
        )
        stone_masses = [
            _decide_boulder_mass(
                rand,
                difficulty,
                min_mass=self.config.min_stone_mass,
                max_mass=self.config.max_stone_mass,
                lighter_is_harder=True,
            )
            for i in range(len(stone_tiles))
        ]

        return (
            spawn_tile,
            food_island_top_left_tile,
            food_island_size,
            food_island_height,
            stone_tiles,
            stone_masses,
            BOX_HEIGHT,
        )

    def add_principal_obstacles(
        self, rand: np.random.Generator, building: Building, obstacle_params: Tuple
    ) -> IndoorTaskParams:
        first_story = building.stories[0]
        main_room = only(first_story.rooms)
        (
            spawn_tile,
            food_island_top_left_tile,
            food_island_size,
            food_island_height,
            stone_tiles,
            stone_masses,
            stone_size,
        ) = obstacle_params
        food_island_position = BuildingTile(*food_island_top_left_tile)
        room_with_island, food_island_tiles = add_food_island(
            main_room, food_island_position, food_island_size, food_island_height
        )

        building.stories[0].rooms[0] = room_with_island

        tile_size = 1
        stone_info = []
        for i, stone_tile in enumerate(stone_tiles):
            floor_elevation = room_with_island.floor_heightmap[stone_tile[1], stone_tile[0]] - DEFAULT_FLOOR_THICKNESS
            stone_location_in_room = np.array(
                [
                    stone_tile[0] + tile_size / 2,
                    floor_elevation + stone_size / 2,
                    stone_tile[1] + tile_size / 2,
                ]
            )
            stone_info.append(StoneInfo(stone_location_in_room, stone_size, stone_masses[i]))

        food_island_center_tile = (
            food_island_position.x + food_island_size // 2,
            food_island_position.z + food_island_size // 2,
        )
        target_location_in_room = np.array(
            [
                food_island_center_tile[0] + tile_size / 2,
                room_with_island.floor_heightmap[food_island_center_tile[1], food_island_center_tile[0]]
                + FOOD_HOVER_DIST,
                food_island_center_tile[1] + tile_size / 2,
            ]
        )

        spawn_location_in_room = np.array(
            [
                spawn_tile[0] + tile_size / 2,
                room_with_island.floor_heightmap[spawn_tile[1], spawn_tile[0]] + AGENT_HEIGHT / 2,
                spawn_tile[1] + tile_size / 2,
            ]
        )

        room_offset = np.array([main_room.position.x, 0, main_room.position.z])
        spawn_location_in_building = local_to_global_coords(np.array(spawn_location_in_room), room_offset)
        target_location_in_building = local_to_global_coords(np.array(target_location_in_room), room_offset)

        stones = []
        for stone_location_in_room, stone_size, stone_mass in stone_info:
            stone_location_in_building = local_to_global_coords(np.array(stone_location_in_room), room_offset)
            stone_location = local_to_global_coords(stone_location_in_building, building.position)
            stones.append(Stone(position=stone_location, size=stone_size, mass=stone_mass))

        entrance_sites = ((first_story.num, main_room.id, tuple(Azimuth)),)
        return building, stones, spawn_location_in_building, target_location_in_building, entrance_sites


class StoneInfo(NamedTuple):
    location: Point3DNP
    size: float
    mass: float



#!/usr/bin/env python3

import pathlib
import statistics
from typing import List
from typing import Tuple

CONFIGS = ["basic", "fancy"]
DRIVERS = ["GLES2", "GLES3"]
NPROCS = ["1", "2", "3", "4", "5"]
WORLDS = {
    "032": "small",
    "064": "medium",
    "220": "large",
    "440": "huge",
}


def get_fps_lines_from_log(config: str, driver: str, world: str, process_count: int, process_id: int) -> List[int]:
    log_path = pathlib.Path(f"./{config}.{driver}.{process_count}.{world}/log.{process_id}.txt")
    fps_lines = [x for x in log_path.read_text().split("\n") if x.startswith("Project FPS")]
    fps_counts = [int(x.split()[2]) for x in fps_lines]
    return fps_counts


def get_performance(config: str, driver: str, world: str) -> Tuple[int, int, int]:
    single_process_fps = max(get_fps_lines_from_log(config, driver, world, 1, 1))
    multi_process_fps_list = []
    multi_process_num_list = []
    for i in range(2, 6):
        fps_means = []
        for j in range(1, i + 1):
            fps_lines = get_fps_lines_from_log(config, driver, world, i, j)
            fps_lines = fps_lines[2:]
            fps_means.append(statistics.mean(fps_lines))
        multi_process_fps_list.append(int(sum(fps_means)))
        multi_process_num_list.append(i)
    multi_process_fps = max(multi_process_fps_list)
    multi_process_num = multi_process_num_list[multi_process_fps_list.index(multi_process_fps)]
    return single_process_fps, multi_process_fps, multi_process_num


def show(value_id: int) -> None:
    for driver in DRIVERS:
        for config in CONFIGS:
            values = []
            for world_id, world_name in WORLDS.items():
                value = get_performance(config, driver, world_id)[value_id]
                values.append(int(value))
            print(
                f"{driver} ({config}) & {values[0]:8,} & {values[1]:8,} & {values[2]:8,} & {values[3]:8,} \\\\"
            )  # script


def main() -> None:
    print("single:")  # script
    show(0)
    print()  # script
    print("multi:")  # script
    show(1)
    print()  # script
    print("nproc:")  # script
    show(2)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import Optional

import attr
import numpy as np

from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.tasks.eat import add_food_tree
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.compositional import CompositeTaskConfig
from avalon.datagen.world_creation.worlds.compositional import ForcedComposition
from avalon.datagen.world_creation.worlds.compositional import create_compositional_task
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class GatherTaskConfig(CompositeTaskConfig):
    task: AvalonTask = AvalonTask.GATHER
    # how many extra fruit trees to create. Just uniform random between these two values (inclusive)
    extra_fruit_tree_count_min: int = 1
    extra_fruit_tree_count_max: int = 5


def generate_gather_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: GatherTaskConfig = GatherTaskConfig(),
    _FORCED: Optional[ForcedComposition] = None,
) -> None:
    world, locations = create_compositional_task(rand, difficulty, task_config, export_config, _FORCED=_FORCED)

    # food being everywhere is probably slightly easier
    is_food_everywhere = select_boolean_difficulty(difficulty, rand, final_prob=0.1)
    if is_food_everywhere:
        island_mask = None
    else:
        island_mask = locations.island

    # are more trees easier or harder? Not sure
    food_tree_count = rand.integers(
        task_config.extra_fruit_tree_count_min, task_config.extra_fruit_tree_count_max, endpoint=True
    )

    # just add some extra foods around. No guarantees about their reachability
    for i in range(food_tree_count):
        tree_position = world.get_safe_point(rand, island_mask=island_mask)
        if tree_position is not None:
            _difficulty, _tool, world = add_food_tree(
                rand, difficulty, world, to_2d_point(tree_position), is_tool_food_allowed=False
            )

    export_world(output_path, rand, world)



import attr

from avalon.agent.common.parse_args import parse_args
from avalon.agent.common.test_envs import TestEnvironmentParams
from avalon.agent.common.trainer import OffPolicyTrainer
from avalon.agent.common.util import setup_new_process
from avalon.agent.dreamer.params import DreamerParams
from avalon.common.log_utils import configure_remote_logger

FRAGMENT_LENGTH = 30


@attr.s(auto_attribs=True, frozen=True)
class DreamerTestEnvParams(DreamerParams):
    total_env_steps: int = 20000
    env_params: TestEnvironmentParams = TestEnvironmentParams(long_episode_length=500)
    rollout_model_update_freq: int = 25
    log_rollout_metrics_every: int = 1
    train_steps: int = 1000  # how often to load checkpoints
    value_target_network_update_freq: int = 25

    batch_size: int = 10
    max_fragment_len: int = FRAGMENT_LENGTH
    min_fragment_len: int = FRAGMENT_LENGTH
    pcont: bool = False
    prioritize_ends: bool = False
    actor_grad: str = "reinforce"
    rssm_hidden_size: int = 32
    deter_size: int = 32
    stoch_size: int = 16
    encoder_mlp_hidden_size: int = 32
    decoder_mlp_hidden_size: int = 32

    worker_managers: int = 1
    num_workers: int = 4
    train_gpu: int = 0
    inference_gpus: tuple[int] = (0,)

    log_freq_hist: int = 10
    log_freq_scalar: int = 1
    log_freq_media: int = 25


def run(params: DreamerParams) -> None:
    trainer = OffPolicyTrainer(params)
    try:
        trainer.start_train_rollouts()
        trainer.train()
    finally:
        trainer.shutdown(finish_wandb_quietly=True)
        del trainer.train_dataloader


if __name__ == "__main__":
    configure_remote_logger(level="INFO")
    setup_new_process()
    default_params = DreamerTestEnvParams()
    default_params = parse_args(default_params)
    run(default_params)



"""
Automatically generates code on both the godot side and python side.

For godot, simply creates the deserialization function because godot is too simple
to have a global registry of classes.

For python, creates attrs types for the types that represent useful objects in godot.

The godot side is purposefully restricted to only deserializing Builder types.
Builder types are those which can actually be created from a config.
Spec types are simply the definitions of fields for the Builders.
THis separation is done so that the configured, created objects can have a type-safe
reference to the stateless Spec class.

Currently python distribution classes are not auto generated, but are included in
godot_base_types.py.  This could be fixed in a future version.
"""

import itertools
import re
from pathlib import Path
from typing import Dict
from typing import Final
from typing import Iterable
from typing import List
from typing import Set
from typing import TypeVar

COMMENTS: Final = {
    "SimSpec": "TODO: should SimSpec still inherit from DataConfigImplementation?",
}

T = TypeVar("T")
_BUILTIN_TYPE_NAMES = ("String", "bool", "int", "float", "Vector2", "Vector3")


def main() -> None:
    generate_godot_code()
    generate_python_code()


def generate_godot_code() -> None:
    pass


def generate_python_code() -> None:
    spec_lines_by_name = load_files("specs")
    distribution_lines_by_name = load_files("utils/distributions")
    externally_defined_types = set(_BUILTIN_TYPE_NAMES + tuple(distribution_lines_by_name.keys()))
    parsed_classes = parse_classes(spec_lines_by_name, externally_defined_types)

    output = []
    output.extend(get_header_lines())

    output.extend(get_const_lines())

    for parsed_class in parsed_classes:
        output.extend(parsed_class.get_lines())

    output_file = Path(__file__).parent.joinpath("godot_generated_types.py")
    print(f"Writing generated types to {output_file}")  # script
    with open(output_file, "w") as out:
        out.write("\n".join(output).strip() + "\n")


def get_const_lines() -> List[str]:
    const_file = Path(__file__).parent.joinpath(f"godot/game/utils/caps/CONST.gd")
    const_lines = []
    with open(const_file, "r") as infile:
        for line in infile.readlines():
            if line.startswith("const "):
                if "[" not in line and "{" not in line:
                    declaration = line.replace(":=", "=").replace("const ", "").strip()
                    variable, value = re.split(r"\s*=\s+", declaration, 1)
                    const_lines.append(f"{variable}: Final = {value}")
    return const_lines + ["", ""]


class ParsedAttribute:
    def __init__(self, name: str, type_name: str, docstring: str, requirements: Set[str], is_optional: bool) -> None:
        self.name = name
        self.type_name = type_name
        self.docstring = docstring
        self.requirements = requirements
        self.is_optional = is_optional

    def get_line(self) -> str:
        if self.is_optional:
            type_name = f"Optional[{self.python_type_name}]"
        else:
            type_name = self.python_type_name
        line = f"    {self.name}: {type_name}"
        if self.docstring:
            line = f"    {self.docstring}\n{line}"
        return line

    @property
    def python_type_name(self) -> str:
        if self.type_name == "String":
            return "str"
        if self.type_name == "Vector2":
            return "AnyVec2"
        if self.type_name == "Vector3":
            return "AnyVec3"
        return self.type_name


class ParsedClass:
    def __init__(self, name: str, base: str, attributes: List[ParsedAttribute]) -> None:
        self.name = name
        self.base = base
        self.attributes = attributes

    def get_lines(self) -> List[str]:
        attribute_lines = [x.get_line() for x in self.attributes]
        if len(attribute_lines) == 0:
            attribute_lines.append("    pass")
        lines = [
            f"@attr.s(auto_attribs=True, hash=True, collect_by_mro=True)",
            f"class {self.name}({self.python_base}):",
            *attribute_lines,
            "",
            "",
        ]
        comment = COMMENTS.get(self.name, None)
        if comment is not None:
            return [f"# {comment}", *lines]
        return lines

    @property
    def python_base(self) -> str:
        if self.name == "SimSpec":
            return "SpecBase, DataConfigImplementation"
        return self.base


def parse_classes(
    lines_by_name: Dict[str, List[str]],
    externally_defined_types: Set[str],
) -> List[ParsedClass]:
    lines_by_name.pop("SpecBase")
    generated_classes = [ParsedClass(name="SpecBase", base="Serializable", attributes=[])]
    required_names_by_class_name = {k: get_requirements(v, externally_defined_types) for k, v in lines_by_name.items()}
    while len(lines_by_name) > 0:
        available_class_names = {x.name for x in generated_classes}
        classes_to_generate = [
            x
            for x in lines_by_name.keys()
            if len(required_names_by_class_name[x].difference(available_class_names)) == 0
        ]
        if len(classes_to_generate) == 0:
            print("Dependencies:")  # script
            for class_name in lines_by_name:
                print(f"    {class_name}: {required_names_by_class_name[class_name]}")  # script
            raise Exception("Some classes do not inherit from SpecBase? " + str(list(lines_by_name.keys())))
        for class_name in sorted(classes_to_generate):
            generated_classes.append(parse_class(class_name, lines_by_name[class_name], externally_defined_types))
            lines_by_name.pop(class_name)
    return generated_classes


def get_requirements(lines: List[str], externally_defined_types: Set[str]) -> Set[str]:
    extends_line = only([x for x in lines if x.startswith("extends")])
    base = extends_line.split(" ")[-1]
    attributes = parse_attributes(lines, externally_defined_types)
    return {base}.union(flatten([x.requirements for x in attributes]))


def parse_attributes(lines: List[str], externally_defined_types: Set[str]) -> List[ParsedAttribute]:
    attributes = []
    next_docstring = []
    for line in lines:
        try:
            if line.startswith("#"):
                next_docstring.append(line)
            elif line.startswith("var "):
                name = line.split(":")[0].replace("var ", "")
                type_name = line.split(":")[1].strip()
                if "=" in type_name:
                    type_name = type_name.split("=")[0].strip()
                requirements = set()
                is_optional = "#optional" in type_name
                if is_optional:
                    type_name = type_name.split("#")[0].strip()
                elif "#" in type_name:
                    type_name = type_name.split("#")[0].strip() + type_name.split("#")[1]
                bare_type_name = type_name.split("[")[0]
                if not is_optional and bare_type_name not in externally_defined_types:
                    requirements.add(bare_type_name)
                docstring = "\n    ".join(next_docstring)
                attributes.append(
                    ParsedAttribute(
                        name=name,
                        type_name=type_name,
                        docstring=docstring,
                        requirements=requirements,
                        is_optional=is_optional,
                    )
                )
                next_docstring = []
        except Exception as e:
            raise Exception(f"Error parsing line: {line}") from e
    return attributes


def parse_class(name: str, lines: List[str], externally_defined_types: Set[str]) -> ParsedClass:
    extends_line = only([x for x in lines if x.startswith("extends")])
    base = extends_line.split(" ")[-1]
    return ParsedClass(name=name, base=base, attributes=parse_attributes(lines, externally_defined_types))


def get_header_lines() -> List[str]:
    return [
        "# GENERATED FILE",
        "# See generate_godot_code.py for details",
        "",
        "from typing import Final",
        "from typing import Optional",
        "",
        "import attr",
        "",
        "from avalon.datagen.godot_base_types import *",
        "",
        "",
    ]


def load_files(file_type_name: str) -> Dict[str, List[str]]:
    lines_by_name: Dict[str, List[str]] = {}
    folder = Path(__file__).parent.joinpath(f"godot/game")
    for spec_file in folder.glob(f"{file_type_name}/*.gd"):
        lines = spec_file.read_text().splitlines()
        lines = [x.rstrip() for x in lines if x.strip()]
        class_name_line = only([x for x in lines if x.startswith("class_name")])
        class_name_from_text = class_name_line.split(" ")[-1]
        lines_by_name[class_name_from_text] = lines
    if len(lines_by_name) == 0:
        raise Exception(f"Found no {file_type_name} files...")
    return lines_by_name


def flatten(iterable: Iterable[Iterable[T]]) -> List[T]:
    return list(itertools.chain(*iterable))


def only(x: Iterable[T]) -> T:
    materialized_x = list(x)
    assert len(materialized_x) == 1
    return materialized_x[0]


if __name__ == "__main__":
    main()



import os
import re
from pathlib import Path
from typing import List
from typing import Optional

import wandb
from loguru import logger
from wandb.apis.public import Run
from wandb.apis.public import Runs

EVAL_TEMP_PATH = "/tmp/avalon_eval"

WANDB_ORGANIZATION = "sourceress"

WANDB_TEST_PROJECT = "test"


class RemoteWandbFileNotFound(Exception):
    pass


class RunDoesNotExist(Exception):
    pass


class ProjectDoesNotExist(Exception):
    pass


def get_wandb_file_by_partial_name(run: Run, file_name: str) -> wandb.apis.public.File:
    name_regex = re.compile(pattern=file_name)
    files: List[wandb.apis.public.File] = [f for f in run.files() if name_regex.search(f.name) is not None]
    if len(files) == 0:
        raise RemoteWandbFileNotFound(f"File that matches `{file_name}` does not exist.")
    if len(files) != 1:
        raise Exception(f"Multiple files found matching {file_name}.")
    return files[0]


def get_wandb_runs_by_tag(project: str, tag: str) -> list[Run]:
    api = wandb.Api()

    if project not in [p.name for p in api.projects()]:
        raise ProjectDoesNotExist()

    runs: Runs = api.runs(path=f"{WANDB_ORGANIZATION}/{project}", filters={"tags": {"$in": [tag]}})
    return list(runs)


def get_wandb_run_by_name(project: str, experiment_name: str) -> Run:
    api = wandb.Api()

    if project not in [p.name for p in api.projects()]:
        raise ProjectDoesNotExist()

    # this will get the most recent run, this could be a potential issue if you have experiments with the same name
    runs: Runs = api.runs(path=f"{WANDB_ORGANIZATION}/{project}", filters={"displayName": experiment_name})
    if len(runs) == 0:
        raise RunDoesNotExist(f"Could not find run `{experiment_name}` in project `{project}`.")
    if len(runs) != 1:
        logger.warning(
            f"INFO: Multiple runs found with the same experiment name {experiment_name}. Using the most recent run."
        )
    run: Run = runs[0]
    return run


def get_wandb_run_by_suggestion_uuid(project: str, uuid: str) -> Run:
    """Set a configuration attribute `uuid` to a unique string, adn this function will fetch that run for you."""
    api = wandb.Api()

    if project not in [p.name for p in api.projects()]:
        raise ProjectDoesNotExist()

    # this will get the most recent run, this could be a potential issue if you have experiments with the same name
    runs: Runs = api.runs(path=f"{WANDB_ORGANIZATION}/{project}", filters={"config.suggestion_uuid": uuid})
    if len(runs) == 0:
        raise RunDoesNotExist(f"Could not find suggestion_uuid `{uuid}` in project `{project}`.")
    if len(runs) != 1:
        logger.warning(f"INFO: Multiple runs found with the same suggestion_uuid {uuid}. Using the most recent run.")
    run: Run = runs[0]
    return run


def get_file_from_wandb(project: str, experiment_name: str, file_name: str) -> wandb.apis.public.File:
    run = get_wandb_run_by_name(project, experiment_name)
    return get_wandb_file_by_partial_name(run, file_name)


def download_wandb_checkpoint_from_run(project: str, experiment_name: str, checkpoint_name: str) -> str:
    file = get_file_from_wandb(project=project, experiment_name=experiment_name, file_name=checkpoint_name)
    path = os.path.join(".", file.name)
    wandb.util.download_file_from_url(path, file.url, wandb.Api().api_key)
    return path


def wandb_ensure_api_key() -> None:
    assert os.getenv("WANDB_API_KEY"), "WANDB_API_KEY not defined, make sure it's included in your bashenv.sh!"


def load_checkpoint_from_wandb_run(run_path: str, filename: str) -> str:
    api = wandb.Api()
    run = api.run(run_path)
    run_root = Path(EVAL_TEMP_PATH) / run_path
    os.makedirs(run_root, exist_ok=True)
    bones_checkpoint_path = wandb.restore(filename, run_path=run_path, replace=True, root=str(run_root))
    assert bones_checkpoint_path is not None, "Could not load checkpoint"
    return bones_checkpoint_path.name


def get_latest_checkpoint_filename(run_path: str, prefix: str = "", suffix: str = "") -> Optional[str]:
    api = wandb.Api()
    run = api.run(run_path)
    checkpoint_filenames: List[str] = [
        file.name for file in run.files() if file.name.startswith(prefix) and file.name.endswith(suffix)
    ]
    checkpoint_filenames = sorted(checkpoint_filenames, key=lambda x: int(x[len(prefix) : -len(suffix)]))
    if len(checkpoint_filenames) == 0:
        return None

    return checkpoint_filenames[-1]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import csv
import datetime
import json
import logging
import os
import time
from typing import Dict
from typing import Optional


def gather_metadata() -> Dict:
    date_start = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
    # Gathering git metadata.
    try:
        import git

        try:
            repo = git.Repo(search_parent_directories=True)
            git_sha = repo.commit().hexsha
            git_data = dict(
                commit=git_sha,
                branch=None if repo.head.is_detached else repo.active_branch.name,
                is_dirty=repo.is_dirty(),
                path=repo.git_dir,
            )
        except git.InvalidGitRepositoryError:
            git_data = None
    except ImportError:
        git_data = None
    # Gathering slurm metadata.
    if "SLURM_JOB_ID" in os.environ:
        slurm_env_keys = [k for k in os.environ if k.startswith("SLURM")]
        slurm_data = {}
        for k in slurm_env_keys:
            d_key = k.replace("SLURM_", "").replace("SLURMD_", "").lower()
            slurm_data[d_key] = os.environ[k]
    else:
        slurm_data = None
    return dict(
        date_start=date_start,
        date_end=None,
        successful=False,
        git=git_data,
        slurm=slurm_data,
        env=os.environ.copy(),
    )


class FileWriter:
    def __init__(
        self,
        xpid: Optional[str] = None,
        xp_args: Optional[dict] = None,
        rootdir: str = "~/logs",
        symlink_to_latest: bool = True,
    ) -> None:
        if not xpid:
            # Make unique id.
            xpid = "{proc}_{unixtime}".format(proc=os.getpid(), unixtime=int(time.time()))
        self.xpid = xpid
        self._tick = 0

        # Metadata gathering.
        if xp_args is None:
            xp_args = {}
        self.metadata = gather_metadata()
        # We need to copy the args, otherwise when we close the file writer
        # (and rewrite the args) we might have non-serializable objects (or
        # other unwanted side-effects).
        self.metadata["args"] = copy.deepcopy(xp_args)
        self.metadata["xpid"] = self.xpid

        formatter = logging.Formatter("%(message)s")
        self._logger = logging.getLogger("logs/out")

        # To stdout handler.
        shandle = logging.StreamHandler()
        shandle.setFormatter(formatter)
        self._logger.addHandler(shandle)
        self._logger.setLevel(logging.INFO)

        rootdir = os.path.expandvars(os.path.expanduser(rootdir))
        # To file handler.
        self.basepath = os.path.join(rootdir, self.xpid)
        if not os.path.exists(self.basepath):
            self._logger.info("Creating log directory: %s", self.basepath)
            os.makedirs(self.basepath, exist_ok=True)
        else:
            self._logger.info("Found log directory: %s", self.basepath)

        if symlink_to_latest:
            # Add 'latest' as symlink unless it exists and is no symlink.
            symlink = os.path.join(rootdir, "latest")
            try:
                if os.path.islink(symlink):
                    os.remove(symlink)
                if not os.path.exists(symlink):
                    os.symlink(self.basepath, symlink)
                    self._logger.info("Symlinked log directory: %s", symlink)
            except OSError:
                # os.remove() or os.symlink() raced. Don't do anything.
                pass

        self.paths = dict(
            msg="{base}/out.log".format(base=self.basepath),
            logs="{base}/logs.csv".format(base=self.basepath),
            fields="{base}/fields.csv".format(base=self.basepath),
            meta="{base}/meta.json".format(base=self.basepath),
        )

        self._logger.info("Saving arguments to %s", self.paths["meta"])
        if os.path.exists(self.paths["meta"]):
            self._logger.warning("Path to meta file already exists. " "Not overriding meta.")
        else:
            self._save_metadata()

        self._logger.info("Saving messages to %s", self.paths["msg"])
        if os.path.exists(self.paths["msg"]):
            self._logger.warning("Path to message file already exists. " "New data will be appended.")

        fhandle = logging.FileHandler(self.paths["msg"])
        fhandle.setFormatter(formatter)
        self._logger.addHandler(fhandle)

        self._logger.info("Saving logs data to %s", self.paths["logs"])
        self._logger.info("Saving logs' fields to %s", self.paths["fields"])
        self.fieldnames = ["_tick", "_time"]
        if os.path.exists(self.paths["logs"]):
            self._logger.warning("Path to log file already exists. " "New data will be appended.")
            # Override default fieldnames.
            with open(self.paths["fields"], "r") as csvfile:
                reader = csv.reader(csvfile)
                lines = list(reader)
                if len(lines) > 0:
                    self.fieldnames = lines[-1]
            # Override default tick: use the last tick from the logs file plus 1.
            with open(self.paths["logs"], "r") as csvfile:
                reader = csv.reader(csvfile)
                lines = list(reader)
                # Need at least two lines in order to read the last tick:
                # the first is the csv header and the second is the first line
                # of data.
                if len(lines) > 1:
                    self._tick = int(lines[-1][0]) + 1

        self._fieldfile = open(self.paths["fields"], "a")
        self._fieldwriter = csv.writer(self._fieldfile)
        self._logfile = open(self.paths["logs"], "a")
        self._logwriter = csv.DictWriter(self._logfile, fieldnames=self.fieldnames)

    def log(self, to_log: Dict, tick: Optional[int] = None, verbose: bool = False) -> None:
        if tick is not None:
            raise NotImplementedError
        else:
            to_log["_tick"] = self._tick
            self._tick += 1
        to_log["_time"] = time.time()

        old_len = len(self.fieldnames)
        for k in to_log:
            if k not in self.fieldnames:
                self.fieldnames.append(k)
        if old_len != len(self.fieldnames):
            self._fieldwriter.writerow(self.fieldnames)
            self._logger.info("Updated log fields: %s", self.fieldnames)

        if to_log["_tick"] == 0:
            self._logfile.write("# %s\n" % ",".join(self.fieldnames))

        if verbose:
            self._logger.info(
                "LOG | %s",
                ", ".join(["{}: {}".format(k, to_log[k]) for k in sorted(to_log)]),
            )

        self._logwriter.writerow(to_log)
        self._logfile.flush()

    def close(self, successful: bool = True) -> None:
        self.metadata["date_end"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        self.metadata["successful"] = successful
        self._save_metadata()

        for f in [self._logfile, self._fieldfile]:
            f.close()

    def _save_metadata(self) -> None:
        with open(self.paths["meta"], "w") as jsonfile:
            json.dump(self.metadata, jsonfile, indent=4, sort_keys=True)



from avalon.common.errors import SwitchError
from avalon.datagen.world_creation.geometry import BuildingTile


def rotate_position(
    position: BuildingTile, story_width: int, story_length: int, degrees: int, tile_like: bool = True
) -> BuildingTile:
    # tile_like=True behaves like np.rot90; tile_like=False rotates directly
    if degrees == 0:
        new_x = position.x
        new_z = position.z
    elif degrees == 90:
        new_x = story_length - position.z
        if tile_like:
            new_x -= 1
        new_z = position.x
    elif degrees == -90:
        new_x = position.z
        new_z = story_width - position.x
        if tile_like:
            new_z -= 1
    elif degrees == 180:
        new_x = story_width - position.x
        new_z = story_length - position.z
        if tile_like:
            new_x -= 1
            new_z -= 1
    else:
        raise SwitchError(degrees)
    return BuildingTile(new_x, new_z)



from pathlib import Path
from typing import Optional
from typing import Tuple

import attr
import numpy as np
from scipy import stats

from avalon.datagen.errors import WorldTooSmall
from avalon.datagen.world_creation.configs.export import ExportConfig
from avalon.datagen.world_creation.configs.task import TaskConfig
from avalon.datagen.world_creation.constants import AGENT_HEIGHT
from avalon.datagen.world_creation.constants import BRIDGE_LENGTH
from avalon.datagen.world_creation.constants import CLIMBING_REQUIRED_HEIGHT
from avalon.datagen.world_creation.constants import HALF_AGENT_HEIGHT_VECTOR
from avalon.datagen.world_creation.constants import MAX_BRIDGE_DIST
from avalon.datagen.world_creation.constants import MIN_BRIDGE_DIST
from avalon.datagen.world_creation.constants import UP_VECTOR
from avalon.datagen.world_creation.constants import AvalonTask
from avalon.datagen.world_creation.entities.tools.log import Log
from avalon.datagen.world_creation.tasks.eat import add_food_tree_for_simple_task
from avalon.datagen.world_creation.utils import to_2d_point
from avalon.datagen.world_creation.worlds.creation import create_world_from_constraint
from avalon.datagen.world_creation.worlds.difficulty import difficulty_variation
from avalon.datagen.world_creation.worlds.difficulty import normal_distrib_range
from avalon.datagen.world_creation.worlds.difficulty import scale_with_difficulty
from avalon.datagen.world_creation.worlds.difficulty import select_boolean_difficulty
from avalon.datagen.world_creation.worlds.export import export_world
from avalon.datagen.world_creation.worlds.obstacles.configure import make_ring
from avalon.datagen.world_creation.worlds.obstacles.height_path import HeightPath
from avalon.datagen.world_creation.worlds.obstacles.height_solution import HeightSolution
from avalon.datagen.world_creation.worlds.types import CompositionalConstraint
from avalon.datagen.world_creation.worlds.utils import add_offsets
from avalon.datagen.world_creation.worlds.world import World
from avalon.datagen.world_creation.worlds.world_locations import WorldLocations


@attr.s(auto_attribs=True, collect_by_mro=True, hash=True)
class BridgeTaskConfig(TaskConfig):
    # how large the gap to bridge should be (ie, how far would you have to jump if you could jump that far)
    # at difficulty 0.0 and 1.0 respectively
    bridge_dist_easy: float = MIN_BRIDGE_DIST
    bridge_dist_hard: float = MAX_BRIDGE_DIST
    # minimum required starting distance between the log and player at difficulty 0.0 and 1.0 respectively
    player_to_log_dist_easy: float = 2.0
    player_to_log_dist_hard: float = 6.0
    # minimum required starting distance between the log and the gap at difficulty 0.0 and 1.0 respectively
    gap_to_log_dist_easy: float = 1.5
    gap_to_log_dist_hard: float = 4.0
    # how wide the gap area is (ie, how much room to the left and right do you have as you are facing across the gap)
    # at difficulty 0.0 and 1.0 respectively
    gap_width_easy: float = 10.0
    gap_width_hard: float = 2.0
    # how wide of an area (in the same sense as gap_width) is flattened. This is required
    # at difficulty 0.0 and 1.0 respectively
    flatten_width_easy: float = 8.0
    flatten_width_hard: float = 4.0


def generate_bridge_task(
    rand: np.random.Generator,
    difficulty: float,
    output_path: Path,
    export_config: ExportConfig,
    task_config: BridgeTaskConfig = BridgeTaskConfig(),
) -> None:
    world, locations, difficulty = create_bridge_obstacle(rand, difficulty, export_config, task_config=task_config)
    world, locations = world.end_height_obstacles(
        locations, is_accessible_from_water=False, is_spawn_region_climbable=False
    )
    world = add_food_tree_for_simple_task(world, locations)
    world = world.add_spawn(rand, difficulty, locations.spawn, locations.goal, is_spawn_height_reset=False)
    export_world(output_path, rand, world)


def create_bridge_obstacle(
    rand: np.random.Generator,
    difficulty: float,
    export_config: ExportConfig,
    constraint: Optional[CompositionalConstraint] = None,
    is_for_carry: bool = False,
    task_config: BridgeTaskConfig = BridgeTaskConfig(),
) -> Tuple[World, WorldLocations, float]:

    is_solved, difficulty = select_boolean_difficulty(difficulty, rand)

    extra_safety_radius = scale_with_difficulty(
        difficulty, task_config.player_to_log_dist_easy, task_config.player_to_log_dist_hard
    )
    desired_gap_distance = scale_with_difficulty(
        difficulty, task_config.bridge_dist_easy, task_config.bridge_dist_hard
    )
    desired_goal_dist = (
        task_config.bridge_dist_easy * 3.0
        + extra_safety_radius * 3.0
        + task_config.bridge_dist_easy * 3.0 * difficulty
    )

    world, locations = create_world_from_constraint(
        stats.norm(desired_goal_dist, desired_goal_dist / 5), rand, difficulty, export_config, constraint
    )

    gap_distance, _warnings = world.get_critical_distance(
        locations, task_config.bridge_dist_easy, desired_gap_distance
    )
    height = scale_with_difficulty(difficulty, CLIMBING_REQUIRED_HEIGHT, BRIDGE_LENGTH)
    inside_item_radius = 0.5
    solution_point_brink_distance = 1.0

    if gap_distance is None:
        raise WorldTooSmall(AvalonTask.BRIDGE, task_config.bridge_dist_easy, locations.get_2d_spawn_goal_distance())

    if is_solved:
        randomization_dist = 0.0
        log_dist = inside_item_radius + (solution_point_brink_distance + gap_distance) / 2
        log_height = height
    else:
        randomization_dist = difficulty_variation(1.0, extra_safety_radius, rand, difficulty)
        log_dist = -difficulty_variation(
            task_config.gap_to_log_dist_easy, task_config.gap_to_log_dist_hard, rand, difficulty
        )
        log_height = 0

    log = Log(position=np.array([log_dist, log_height, 0.0]))

    ring_config = make_ring(
        rand,
        difficulty,
        world,
        locations,
        gap_distance,
        constraint=constraint,
        height=-height,
        traversal_width=normal_distrib_range(
            task_config.gap_width_easy, task_config.gap_width_hard, 1.0, rand, difficulty
        ),
        is_inside_climbable=True,
        is_outside_climbable=False,
        dual_solution=HeightSolution(
            inside_items=tuple(add_offsets([log])),
            inside_item_randomization_distance=randomization_dist,
            inside_item_radius=inside_item_radius,
            paths=(
                HeightPath(
                    is_solution_flattened=True,
                    is_height_affected=False,
                    width=normal_distrib_range(
                        task_config.flatten_width_easy, task_config.flatten_width_hard, 1.0, rand, difficulty
                    ),
                ),
            ),
            solution_point_brink_distance=solution_point_brink_distance,
        ),
        extra_safety_radius=extra_safety_radius,
        probability_of_centering_on_spawn=0.0 if is_for_carry else 0.5,
    )
    world = world.add_height_obstacle(rand, ring_config, locations.island)

    # if this is non-compositional (ie, normal) task, then make sure we didn't accidentally overlap the spawn and log
    if constraint is None:
        # reset the spawn no matter what:
        spawn_location = locations.spawn.copy()
        idx = world.map.point_to_index(to_2d_point(spawn_location))
        height_at_location = world.map.Z[idx]
        spawn_location[1] = height_at_location + HALF_AGENT_HEIGHT_VECTOR[1] * 1.1

        # fix up the spawn if too close to the log
        item = world.items[-1]
        assert isinstance(item, Log), "Huh, only expected a single item to be spawned in this case..."
        log = item
        spawn_log_dist = np.linalg.norm(locations.spawn - log.position)
        safety_margin = 0.5
        if spawn_log_dist < (BRIDGE_LENGTH / 2.0) + safety_margin:
            spawn_location = spawn_location + UP_VECTOR * (AGENT_HEIGHT + safety_margin)

        # set the spawn
        locations = attr.evolve(locations, spawn=spawn_location)

    return world, locations, difficulty
